/* oe_cholesky.c -- Cholesky decomposition functions for optimal
   estimation library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <math.h>

#include "oe.h"

/* Perform Cholesky decomposition on symmetric, positive-definite
   matrix "A", returning the result (the matrix L) in the lower half
   of matrix A, except for the diagonals of L which are returned in
   the vector "diagonal". This function is an adaption from the one in
   Numerical Recipies (second edition). */
int
oe_cholesky_decomposition(OEmatrix *A, OEvector *diagonal)
{
  OEreal **A_data = A->data;
  OEreal sum;
  int n = A->n;
  int i, j, k;

  if (n != A->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Attempt to Cholesky-decompose a non-square matrix");
  }

  OE_CHECK(oe_reset_vector(diagonal, n));

  for (i = 0; i < n; i++) {
    for (j = i; j < n; j++) {
      sum = A_data[i][j];
      for (k = i-1; k >= 0; k--) {
	sum -= A_data[i][k] * A_data[j][k];
      }
      if (i == j) {
	if (sum <= 0.0) {
	  OE_THROW(OE_MATRIX_ERROR, 
	    "Attempt to Cholesky-decompose a non positive definite matrix");
	  return OE_FAILURE;
	}
	diagonal->data[i] = sqrt(sum);
      }
      else {
	A_data[j][i] = sum / diagonal->data[i];
      }
    }
  }
  return OE_SUCCESS;
}

/* Interpret "A" and "diagonal" as the output of
   oe_cholesky_decomposition() and use them to solve the system of
   equations A*x=b (where here A is the matrix before its
   decomposition). The vector b is overwritten with the result x. */
int
oe_cholesky_substitution(OEmatrix *A, OEvector *diagonal, OEvector *b)
{
  OEreal** A_data = A->data;
  OEreal sum;
  int i, k;
  int n = A->n;

  if (n != A->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Attempt to backsubstitute a non-square matrix");
  }

  if (n != diagonal->n) {
    OE_THROW(OE_DIMENSION_ERROR, "Invalid diagonal vector");
  }

  if (n != b->n) {
    OE_THROW(OE_DIMENSION_ERROR, 
    "Vector and matrix have incompatible dimensions in Cholesky substitution");
    return OE_FAILURE;
  }

  for (i = 0; i < n; i++) {
    sum = b->data[i];
    for (k = i-1; k >= 0; k--) {
      sum -= A_data[i][k] * b->data[k];
    }
    b->data[i] = sum/diagonal->data[i];
  }
  for (i = n-1; i >= 0; i--) {
    sum = b->data[i];
    for (k = i+1; k < n; k++) {
      sum -= A_data[k][i] * b->data[k];
    }
    b->data[i] = sum/diagonal->data[i];
  }
  return OE_SUCCESS;
}

/* Interpret A as a matrix returned by oe_cholesky_decomposition (the
   lower part containing L, except for the diagonals of L which are in
   "diagonal"), and calculate the inverse of L, returning the result
   in the lower part of A (including the diagonals this time). This
   code has not been tested. */ 
int
oe_cholesky_invert_lower(OEmatrix *A, OEvector *diagonal)
{
  OEreal **A_data = A->data;
  OEreal sum;
  int i, j, k;
  int n = A->n;

  if (n != A->m) {
    OE_THROW(OE_DIMENSION_ERROR, "Cholesky matrix is not square");
  }
  else if (n != diagonal->n) {
    OE_THROW(OE_DIMENSION_ERROR, "Dimensions of Cholesky matrix and associated vector of diagonals do not match");
  }

  for (i = 0; i < n; i++) {
    A_data[i][i] = 1.0/diagonal->data[i];
    for (j = i+1; j < n; j++) {
      sum = 0.0;
      for (k = i; k < j; k++) {
	sum -= A_data[j][k] * A_data[k][i];
      }
      A_data[j][i] = sum/diagonal->data[j];
    }
  }
  return OE_SUCCESS;
}
