/* oe_compound.h -- Compound matrix operations for optimal estimation
   library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "oe.h"

/* Intermediate data for sparse matrix operations */
static OEintmatrix _index = OE_NEW_MATRIX;
static OEintvector _num_indeces = OE_NEW_VECTOR;

/* Taking K to be a non-square Jacobian and invS to be the inverse of
   a diagonal covariance matrix, multiply invS by K and return it in
   invS_K, then multiply the transpose of K by invS_K and increment
   the Hessian "sumA" with the result. Matrices are assumed to be
   non-sparse so this operation is slow. */
int
oe_KT_invs_K(OEmatrix *K, OEvector *invS, OEmatrix *sumA, OEmatrix *invS_K)
{
  int m = K->m;
  int n = K->n;
  int i, j, h;
  
  if (m != invS->n) {
    OE_THROW(OE_DIMENSION_ERROR, 
	     "Dimensions of input matrix and vector do not match");
  }
  if (n != sumA->n || n != sumA->m) {
    OE_THROW(OE_DIMENSION_ERROR,
     "Hessian output matrix dimensions do not match Jacobian input matrix");
  }
  if (m != invS_K->m || n != invS_K->n) {
    OE_THROW(OE_DIMENSION_ERROR,
     "invS_K product matrix dimensions do not match Jacobian input matrix");
  }

  /* Set invS_K = invS*K */
  for (j = 0; j < m; j++) {
    for (i = 0; i < n; i++) {
      invS_K->data[j][i] = invS->data[j] * K->data[j][i];
    }
  }

  /* Add the K'*invS_K term to sumA */
  for (j = 0; j < m; j++) {
    for (i = 0; i < n; i++) {
      for (h = 0; h <= i; h++) {
	sumA->data[i][h] += K->data[j][i] * invS_K->data[j][h];
      }
    }
  }
  /* Ensure that sumA is symmetric */
  for (i = 0; i < n; i++) {
    for (h = i+1; h < n; h++) {
      sumA->data[i][h] = sumA->data[h][i];
    }
  }
  return OE_SUCCESS;
}

/* As oe_KT_invs_K() but optimised for sparse K. It works by first
   forming a list of the non-zero elements of each column of K, then
   using this to reduce the number of floating-point
   multiplications. */
int
oe_sparse_KT_invs_K(OEmatrix *K, OEvector *invS, OEmatrix *sumA,
		    OEmatrix *invS_K)
{
  int m = K->m;
  int n = K->n;
  int i, j, h;
  int *index;
  int num;

  if (m != invS->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Dimensions of input matrix and vector do not match");
  }
  if (n != sumA->n || n != sumA->m) {
    OE_THROW(OE_DIMENSION_ERROR,
     "Hessian output matrix dimensions do not match Jacobian input matrix");
  }
  if (m != invS_K->m || n != invS_K->n) {
    OE_THROW(OE_DIMENSION_ERROR,
     "invS_K product matrix dimensions do not match Jacobian input matrix");
  }

  OE_CHECK(oe_reset_int_matrix(&_index, n, m));
  OE_CHECK(oe_reset_int_vector(&_num_indeces, n));
  
  /* Form a list of the non-zero elements of each column of K */
  for (i = 0; i < n; i++) {
    int sum = 0;
    for (j = 0; j < m; j++) {
      if (K->data[j][i] != 0.0) {
	_index.data[i][sum++] = j;
      }
    }
    /* Save the number of non-zero elements of this column */
    _num_indeces.data[i] = sum;
  }

  oe_zero_matrix(invS_K);
  /* Set invS_K = invS*K */
  for (i = 0; i < n; i++) {
    for (j = 0; j < _num_indeces.data[i]; j++) {
      h = _index.data[i][j];
      invS_K->data[h][i] = invS->data[h] * K->data[h][i];
    }
  }

  /* Add the K'*invS_K term to sumA */
  for (i = 0; i < n; i++) {
    for (h = 0; h <= i; h++) {
      /* We are essentially multiplying one column of K with another -
	 this is optimised by finding which of the two has the least
	 number of non-zero elements and using this list of elements. */
      if (_num_indeces.data[i] < _num_indeces.data[h]) {
        index = _index.data[i];
        num = _num_indeces.data[i];
      }
      else {
        index = _index.data[h];
        num = _num_indeces.data[h];
      }
      for (j = 0; j < num; j++) {
        sumA->data[i][h] += K->data[index[j]][i] * invS_K->data[index[j]][h];
      }
    }
  }

  /* Ensure that sumA is symmetric */
  for (i = 0; i < n; i++) {
    for (h = i+1; h < n; h++) {
      sumA->data[i][h] = sumA->data[h][i];
    }
  }
  return OE_SUCCESS;
}

int
oe_sparse_W_invA_WT_diagonal(OEmatrix *W, OEmatrix *invA, OEvector *v_out)
{
  int m = W->m;
  int n = W->n;
  int i;

  if (invA->m != invA->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Input inverse Hessian matrix is not square");
  }
  else if (n != invA->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Inner dimensions of input weighting and inverse Hessian matrices do not match");
  }
  else if (m != v_out->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Height of input weighting matrix does not match length of output vector");
  }

  OE_CHECK(oe_reset_int_vector(&_num_indeces, n));

  for (i = 0; i < m; i++) {
  /* Form a list of the non-zero elements of each row of W */
    OEreal result = 0.0;
    int sum = 0;
    int j, k, l;
    for (j = 0; j < n; j++) {
      if (W->data[i][j] != 0.0) {
	_num_indeces.data[sum++] = j;
      }
    }
    for (k = 0; k < sum; k++) {
      for (l = 0; l < sum; l++) {
	result += invA->data[_num_indeces.data[k]][_num_indeces.data[l]]
	  * W->data[i][_num_indeces.data[l]]
	  * W->data[i][_num_indeces.data[k]];
      }
    }
    v_out->data[i] += result;
  }
  return OE_SUCCESS;  
}
