/* oe_errcov.h -- Generate idealised error covariance matrices for
   optimal estimation library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <math.h>
#include "oe.h"

/* Produce an idealized error covariance matrix consisting as a
   constant value on the diagonal ("diagval") and an exponential
   drop-off to each side with a specified decorrelation distance (in
   units of the spacing of the grid); the result is a symmetric
   Toeplitz matrix. */
int
oe_exponential_errcov(int n, OEreal diagval, OEreal indexdecorr,
		      OEmatrix *M)
{
  int i, j;
  OEreal factor = -1.0/indexdecorr;
  OE_CHECK(oe_reset_matrix(M, n, n));
  for (i = 0; i < n; i++) {
    for (j = i; j < n; j++) {
      M->data[i][j] = M->data[j][i] = diagval * exp(factor * (j-i));
    }
  }
  return OE_SUCCESS;
}

/* Produce an idealized error covariance matrix consisting as a
   constant value on the diagonal ("diagval") and a Gaussian drop-off
   to each side with a specified decorrelation distance (in units of
   the spacing of the grid); the result is a symmetric Toeplitz
   matrix. */
int
oe_gaussian_errcov(int n, OEreal diagval, OEreal indexdecorr,
		   OEmatrix *M)
{
  int i, j;
  OEreal factor = -0.5/(indexdecorr*indexdecorr);
  OE_CHECK(oe_reset_matrix(M, n, n));
  for (i = 0; i < n; i++) {
    for (j = i; j < n; j++) {
      M->data[i][j] = M->data[j][i] = diagval * exp(factor * (j-i)*(j-i));
    }
  }
  return OE_SUCCESS;
}
