/* oe_netcdf.c -- Interface between oe structures and NetCDF datasets

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>

#include "oe_netcdf.h"

/* See oe_netcdf.h for how "ncstatus" is used */
int ncstatus;

/* Get the "missing value" for an existing NetCDF variable (looking
   for the attribute "_FillValue" before "missing_value"), returning
   OE_FAILURE if neither of these two is found, or if the attribute is
   of the wrong type or length */
static
int
get_missing_value(int ncid, int varid, OEreal *missing_value)
{
  size_t length;
  OEreal value;
  char *name = "_FillValue";
  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    name = "missing_value";
    if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
      return OE_FAILURE;
    }
  }
  if (length > 1) {
    return OE_FAILURE;
  }
  if (nc_get_att_double(ncid, varid, name, &value) != NC_NOERR) {
    return OE_FAILURE;
  }
  else {
    *missing_value = (double) value;
  }
  return OE_SUCCESS;
}

/* Read a scalar NetCDF variable into "value" */
int
oe_read_nc_scalar(int ncid, char *varname, OEreal *value)
{
  int varid, ndims;
  size_t index = 0;
  NC_CHECK(nc_inq_varid(ncid, varname, &varid));
  NC_CHECK(nc_inq_varndims(ncid, varid, &ndims));
  if (ndims > 1) {
    OE_THROW(OE_DIMENSION_ERROR, "NetCDF variable has more than one dimension");
  }
  NC_CHECK(nc_get_var1_double(ncid, varid, &index, value));
  return OE_SUCCESS;

}

/* Read vector named "varname" from a NetCDF dataset (of type float or
   double) into OEvector "v"; an error is thrown if the variable is
   not present or of the wrong type */
int
oe_read_nc_vector(int ncid, char *varname, OEvector *v)
{
  OEreal missing_value;
  int varid, ndims, dimid;
  size_t len;
 
  NC_CHECK(nc_inq_varid(ncid, varname, &varid));
  NC_CHECK(nc_inq_varndims(ncid, varid, &ndims));
  if (ndims != 1) {
    OE_THROW(OE_DIMENSION_ERROR, "NetCDF variable is not a 1D vector");
  }
  NC_CHECK(nc_inq_vardimid(ncid, varid, &dimid));
  NC_CHECK(nc_inq_dimlen(ncid, dimid, &len));
  OE_CHECK(oe_reset_vector(v, len));
  NC_CHECK(nc_get_var_double(ncid, varid, v->data));
  if (get_missing_value(ncid, varid, &missing_value) == OE_SUCCESS) {
    v->missing_value = missing_value;
    if (oe_verbose) {
      fprintf(stderr, "Loaded %s (vector with missing_value = %g)\n",
	      varname, missing_value);
    }
  }
  else if (oe_verbose) {
    fprintf(stderr, "Loaded %s (vector with no missing_value)\n", varname);
  }
  return OE_SUCCESS;
}

/* Read matrix named "varname" from a NetCDF dataset (of type float or
   double) into OEmatrix "M"; an error is thrown if the variable is
   not present or of the wrong type */
int
oe_read_nc_matrix(int ncid, char *varname, OEmatrix *M)
{
  OEreal missing_value;
  int varid, ndims, dimid[2];
  size_t m, n;
  int i;
  size_t start[2] = {0, 0}, count[2] = {1, 1};

  NC_CHECK(nc_inq_varid(ncid, varname, &varid));
  NC_CHECK(nc_inq_varndims(ncid, varid, &ndims));
  if (ndims != 2) {
    OE_THROW(OE_DIMENSION_ERROR, "NetCDF variable is not a 2D matrix");
  }
  NC_CHECK(nc_inq_vardimid(ncid, varid, dimid));
  NC_CHECK(nc_inq_dimlen(ncid, dimid[0], &m));
  NC_CHECK(nc_inq_dimlen(ncid, dimid[1], &n));
  OE_CHECK(oe_reset_matrix(M, m, n));

  count[1] = n;
  for (i = 0; i < m; i++) {
    start[0] = i;
    NC_CHECK(nc_get_vara_double(ncid, varid, start, count, M->data[i]));
  }
  if (get_missing_value(ncid, varid, &missing_value) == OE_SUCCESS) {
    M->missing_value = missing_value;
    if (oe_verbose) {
      fprintf(stderr, "Loaded %s (matrix with missing_value = %g)\n",
	      varname, missing_value);
    }
  }
  else if (oe_verbose) {
    fprintf(stderr, "Loaded %s (matrix no missing_value)\n",
	    varname);
  }
  return OE_SUCCESS;
}

/* Read slice "islice" of a 2D NetCDF field into OEvector "v" */
int
oe_read_nc_slice(int ncid, char *varname, int islice, OEvector *v)
{
  OEreal missing_value;
  int varid, ndims, dimid[2];
  size_t m, n;
  size_t start[2] = {0, 0}, count[2] = {1, 1};

  NC_CHECK(nc_inq_varid(ncid, varname, &varid));
  NC_CHECK(nc_inq_varndims(ncid, varid, &ndims));
  if (ndims != 2) {
    OE_THROW(OE_DIMENSION_ERROR, "NetCDF variable is not a 2D matrix");
  }
  NC_CHECK(nc_inq_vardimid(ncid, varid, dimid));
  NC_CHECK(nc_inq_dimlen(ncid, dimid[0], &m));
  NC_CHECK(nc_inq_dimlen(ncid, dimid[1], &n));

  if (islice < 0 || islice >= m) {
    OE_THROW(OE_DIMENSION_ERROR, "Requested slice out of range");
  }

  OE_CHECK(oe_reset_vector(v, n));

  count[1] = n;
  start[0] = islice;
  NC_CHECK(nc_get_vara_double(ncid, varid, start, count, v->data));
  if (get_missing_value(ncid, varid, &missing_value) == OE_SUCCESS) {
    v->missing_value = missing_value;
    if (oe_verbose) {
      fprintf(stderr, "Loaded slice %d of %s (missing_value = %g)\n",
	      islice, varname, missing_value);
    }
  }
  else if (oe_verbose) {
    fprintf(stderr, "Loaded slice %d of %s (no missing_value)\n",
	    islice, varname);
  }
  return OE_SUCCESS;

}

/* Read slice "islice" of a 2D NetCDF integer field into OEvector "v" */
int
oe_read_nc_int_slice(int ncid, char *varname, int islice, OEintvector *v)
{
  int varid, ndims, dimid[2];
  size_t m, n;
  size_t start[2] = {0, 0}, count[2] = {1, 1};

  NC_CHECK(nc_inq_varid(ncid, varname, &varid));
  NC_CHECK(nc_inq_varndims(ncid, varid, &ndims));
  if (ndims != 2) {
    OE_THROW(OE_DIMENSION_ERROR, "NetCDF variable is not a 2D matrix");
  }
  NC_CHECK(nc_inq_vardimid(ncid, varid, dimid));
  NC_CHECK(nc_inq_dimlen(ncid, dimid[0], &m));
  NC_CHECK(nc_inq_dimlen(ncid, dimid[1], &n));

  if (islice < 0 || islice >= m) {
    OE_THROW(OE_DIMENSION_ERROR, "Requested slice out of range");
  }

  OE_CHECK(oe_reset_int_vector(v, n));

  count[1] = n;
  start[0] = islice;
  NC_CHECK(nc_get_vara_int(ncid, varid, start, count, v->data));
  if (oe_verbose) {
    fprintf(stderr, "Loaded int slice %d of %s\n", islice, varname);
  }

  return OE_SUCCESS;

}

/* Put OEmatrix "M" into a NetCDF field with id "fieldid" */
int
oe_put_nc_matrix(int ncid, int fieldid, OEmatrix *M)
{
  size_t start[2], count[2];
  int i;

  start[1] = 0;
  count[0] = 1;
  count[1] = M->n;

  for (i = 0; i < M->m; i++) {
    start[0] = i;
    NC_CHECK(nc_put_vara_double(ncid, fieldid, start, count, M->data[i]));
  }
  return OE_SUCCESS;
}

/* Put OEvector "v" into 2D NetCDF field with id "fieldid", as slice
   "islice" */
int
oe_put_nc_slice(int ncid, int fieldid, int islice, OEvector *v)
{
  size_t start[2], count[2];

  start[0] = islice;
  start[1] = 0;
  count[0] = 1;
  count[1] = v->n;

  NC_CHECK(nc_put_vara_double(ncid, fieldid, start, count, v->data));

  return OE_SUCCESS;
}
