/* oe_prepare.h -- Prepare matrices for the solver in optimal
   estimation library

   Copyright (C) 2005 Robin Hogan <r.j.hogan@reading.ac.uk> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include "oe.h"

/* Internal matrices */
static OEmatrix _M = OE_NEW_MATRIX;
static OEvector _xa[OE_MAX_PRIOR_CONSTRAINTS]
  = {OE_NEW_VECTOR,OE_NEW_VECTOR,OE_NEW_VECTOR,OE_NEW_VECTOR};
static OEmatrix _inv_Sa[OE_MAX_PRIOR_CONSTRAINTS]
  = {OE_NEW_MATRIX,OE_NEW_MATRIX,OE_NEW_MATRIX,OE_NEW_MATRIX};
static OEvector _inv_sa[OE_MAX_PRIOR_CONSTRAINTS]
  = {OE_NEW_VECTOR,OE_NEW_VECTOR,OE_NEW_VECTOR,OE_NEW_VECTOR};
static OEmatrix _H = OE_NEW_MATRIX;

/* Information about constraints */
static int is_a_priori[OE_MAX_PRIOR_CONSTRAINTS] = {0, 0, 0, 0};
static int a_priori_is_diagonal[OE_MAX_PRIOR_CONSTRAINTS] = {0, 0, 0, 0};
static int is_smoothness_constraint = 0;

/* Print information about each of the a-priori constraints */
int
oe_print_a_priori_status(FILE *file)
{
  int i;
  for (i = 0; i < OE_MAX_PRIOR_CONSTRAINTS; i++) {
    fprintf(file, "Constraint %d: ", i);
    if (is_a_priori[i]) {
      fprintf(file, "diagonal=%d; size=%d/%d,%d\n", a_priori_is_diagonal[i],
	     _inv_sa[i].n, _inv_Sa[i].m, _inv_Sa[i].n);
    }
    else {
      fprintf(file, "none\n");
    }
  }
  return OE_SUCCESS;
}

/* Register an a-priori estimate of the solution (xa) and its
   associated covariance matrix (Sa); the former is stored internally
   whereas the matrix inverse of the latter is stored. These are then
   used in oe_prepare().  */
int
oe_register_a_priori_i(int i, OEvector *xa, OEmatrix *Sa)
{
  if (i >= OE_MAX_PRIOR_CONSTRAINTS) {
    OE_THROW(OE_FAILURE,
	     "Number of a priori constraint out of range");
  }
  else if (xa == NULL || Sa == NULL) {
    is_a_priori[i] = 0;
    return OE_SUCCESS;
  }
  else if (Sa->n != Sa->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "A priori error covariance matrix is not square");
  }
  else if (Sa->n != xa->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Size mismatch between a priori and its error covariance matrix");
  }
  OE_CHECK(oe_copy_vector(xa, &_xa[i]));

  //  OE_CHECK(oe_copy_matrix(Sa, &_inv_Sa[i]));
  //  OE_CHECK(oe_save_matrix("inv_Sa.dat", &_inv_Sa[i], "% inv_Sa"));
  //  exit(1);
  
  OE_CHECK(oe_reset_matrix(&_inv_Sa[i], Sa->n, Sa->n));
  //  OE_CHECK(oe_invert(Sa, &_inv_Sa[i]));
  OE_CHECK(oe_invert_symmetric(Sa, &_inv_Sa[i]));
  
  is_a_priori[i] = 1;
  a_priori_is_diagonal[i] = 0;

  return OE_SUCCESS;
}

/* This can be used if only one a-priori is required */
int
oe_register_a_priori(OEvector *xa, OEmatrix *Sa)
{
  OE_CHECK(oe_register_a_priori_i(0, xa, Sa));
  return OE_SUCCESS;
}

/* Register an a-priori estimate of the solution (xa) and its
   associated inverse covariance matrix (Sa); both are stored
   internally. */
int
oe_register_a_priori_invs_i(int i, OEvector *xa, OEmatrix *invSa)
{
  if (i >= OE_MAX_PRIOR_CONSTRAINTS) {
    OE_THROW(OE_FAILURE,
	     "Number of a priori constraint out of range");
  }
  else if (xa == NULL || invSa == NULL) {
    is_a_priori[i] = 0;
    return OE_SUCCESS;
  }
  else if (invSa->n != invSa->m) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Inverse a priori error covariance matrix is not square");
  }
  else if (invSa->n != xa->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Size mismatch between a priori and its inverse error covariance matrix");
  }
  OE_CHECK(oe_copy_vector(xa, &_xa[i]));

  //  OE_CHECK(oe_copy_matrix(Sa, &_inv_Sa[i]));
  //  OE_CHECK(oe_save_matrix("inv_Sa.dat", &_inv_Sa[i], "% inv_Sa"));
  //  exit(1);
  //  OE_CHECK(oe_invert_symmetric(&_inv_Sa[i]));
  
  OE_CHECK(oe_copy_matrix(invSa, &_inv_Sa[i]));
  
  is_a_priori[i] = 1;
  a_priori_is_diagonal[i] = 0;

  return OE_SUCCESS;
}

/* This can be used if only one a-priori is required */
int
oe_register_a_priori_invs(OEvector *xa, OEmatrix *invSa)
{
  OE_CHECK(oe_register_a_priori_invs_i(0, xa, invSa));
  return OE_SUCCESS;
}

/* Register an a-priori estimate of the solution (xa) and the diagonal
   elements of its covariance matrix (sa), both of which are stored
   internally. */
int
oe_register_diagonal_a_priori_i(int i, OEvector *xa, OEvector *sa)
{
  int j;
  if (i >= OE_MAX_PRIOR_CONSTRAINTS) {
    OE_THROW(OE_FAILURE,
	     "Number of a priori constraint out of range");
  }
  else if (xa == NULL || sa == NULL) {
    is_a_priori[i] = 0;
    return OE_SUCCESS;
  }
  else if (sa->n != xa->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Size mismatch between a priori and its error variance vector");
  }
  OE_CHECK(oe_copy_vector(xa, &_xa[i]));
  OE_CHECK(oe_reset_vector(&_inv_sa[i], sa->n));

  is_a_priori[i] = 1;
  a_priori_is_diagonal[i] = 1;
  for (j = 0; j < sa->n; j++) {
    if (sa->data[j] == 0.0) {
      is_a_priori[i] = 0;
      OE_THROW(OE_MATRIX_ERROR,
	       "A priori error variance vector contains zeros");
      return OE_FAILURE;
    }
    _inv_sa[i].data[j] = 1.0/sa->data[j];
  }
  return OE_SUCCESS;
}

/* This can be used if only one a-priori is required */
int
oe_register_diagonal_a_priori(OEvector *xa, OEvector *sa)
{
  OE_CHECK(oe_register_diagonal_a_priori_i(0, xa, sa));
  return OE_SUCCESS;
}


/* Register an a-priori estimate of the solution (xa) and the diagonal
   elements of the inverse of its covariance matrix (inv_sa), both of
   which are stored internally. */
int
oe_register_diagonal_a_priori_invs_i(int i, OEvector *xa, OEvector *inv_sa)
{
  if (i >= OE_MAX_PRIOR_CONSTRAINTS) {
    OE_THROW(OE_FAILURE,
	     "Number of a priori constraint out of range");
  }
  else if (xa == NULL || inv_sa == NULL) {
    is_a_priori[i] = 0;
    return OE_SUCCESS;
  }
  else if (inv_sa->n != xa->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Size mismatch between a priori and its error variance vector");
  }
  OE_CHECK(oe_copy_vector(xa, &_xa[i]));
  OE_CHECK(oe_copy_vector(inv_sa, &_inv_sa[i]))

  is_a_priori[i] = 1;
  a_priori_is_diagonal[i] = 1;

  return OE_SUCCESS;
}

/* This can be used if only one a-priori is required */
int
oe_register_diagonal_a_priori_invs(OEvector *xa, OEvector *inv_sa)
{
  OE_CHECK(oe_register_diagonal_a_priori_invs_i(0, xa, inv_sa));
  return OE_SUCCESS;
}

/* Reset the smoothness constraint matrix to n*n, set the contents to
   zero and (until oe_register_smoothness_constraint()) is called,
   turn off the switch that uses it in oe_prepare(). Note that
   oe_prepare() will throw an error if n is not the same as the size
   of the state vector. */
int
oe_reset_smoothness_constraint(int n)
{
  if (n <= 0) {
    is_smoothness_constraint = 0;
    return OE_SUCCESS;
  }
  OE_CHECK(oe_reset_matrix(&_H, n, n));

  /* Note that the smoothness constraint is only activated when the
     oe_register_smoothness_constraint() function is called */
  is_smoothness_constraint = 0;
  OE_CHECK(oe_zero_matrix(&_H));
  return OE_SUCCESS;
}

/* Return a pointer to the smoothness constraint matrix held
   internally */
int
oe_get_smoothness_constraint(OEmatrix **H)
{
  if (is_smoothness_constraint) {
    *H = &_H;
    return OE_SUCCESS;
  }
  else {
    OE_THROW(OE_FAILURE, "Smoothness constraint has not been set");
  }
}

/* Add a smoothness constraint to the elements of the state vector
   between n1 and n2 inclusive, with a specified smoothness
   factor. The matrix will be of the Twomey type as described by
   Rodgers.  */
int
oe_register_smoothness_constraint(OEreal factor, int n1, int n2)
{
  int i;
  if (n1 >= n2 || n2 >= _H.n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Smoothness constraint parameters out of range");
  }
  else if (n2 < n1+2) {
    /* Cannot apply a smoothness constraint to fewer than 3 points*/
    return OE_SUCCESS;
  }
  is_smoothness_constraint = 1;
  /* Do the ends */
  _H.data[n1][n1] += factor;
  _H.data[n2][n2] += factor;
  _H.data[n1][n1+2] += factor;
  _H.data[n2][n2-2] += factor;
  _H.data[n1][n1+1] -= 2.0 * factor;
  _H.data[n2][n2-1] -= 2.0 * factor;
  if (n2 == n1+2) {
    /* Special case if only three points */
    _H.data[n1+1][n1] -= 2.0 * factor; 
    _H.data[n1+1][n1+2] -= 2.0 * factor;
    _H.data[n1+1][n1+1] += 4.0 * factor;
  }
  else {
    _H.data[n1+1][n1] -= 2.0 * factor; 
    _H.data[n2-1][n2] -= 2.0 * factor;
    _H.data[n1+1][n1+1] += 5.0 * factor;
    _H.data[n2-1][n2-1] += 5.0 * factor;
    _H.data[n1+1][n1+2] -= 4.0 * factor;
    _H.data[n2-1][n2-2] -= 4.0 * factor;
    _H.data[n1+1][n1+3] += factor;
    _H.data[n2-1][n2-3] += factor;
    for (i = n1+2; i <= n2-2; i++) {
      _H.data[i][i] += 6.0 * factor;
      _H.data[i][i-1] -= 4.0 * factor;
      _H.data[i][i+1] -= 4.0 * factor;
      _H.data[i][i-2] += factor;
      _H.data[i][i+2] += factor;
    }
  }
  return OE_SUCCESS;
}

/* Add a smoothness constraint to the elements of the state vector
   between n1 and n2 inclusive, with a specified smoothness
   factor. The matrix will be similar to the Twomey type as described
   by Rodgers, except that the edges are excluded. This results in
   behaviour as if there were zero elements beyond the ends of the
   part of the state vector being smoothed, such that the smooth
   solution will tend to be lower at the edges. */
int
oe_register_pinned_smoothness_constraint(OEreal factor, int n1, int n2)
{
  int i;
  if (n1 >= n2 || n2 >= _H.n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Smoothness constraint parameters out of range");
  }
  else if (n2 < n1+2) {
    /* Cannot apply a smoothness constraint to fewer than 3 points*/
    return OE_SUCCESS;
  }
  is_smoothness_constraint = 1;
  /* Do the ends */
  _H.data[n1][n1] += 5.0 * factor;
  _H.data[n2][n2] += 5.0 * factor;
  _H.data[n1][n1+2] += factor;
  _H.data[n2][n2-2] += factor;
  _H.data[n1][n1+1] -= 4.0 * factor;
  _H.data[n2][n2-1] -= 4.0 * factor;
  if (n2 == n1+2) {
    /* Special case if only three points */
    _H.data[n1+1][n1] -= 4.0 * factor; 
    _H.data[n1+1][n1+2] -= 4.0 * factor;
    _H.data[n1+1][n1+1] += 6.0 * factor;
  }
  else {
    _H.data[n1+1][n1] -= 4.0 * factor; 
    _H.data[n2-1][n2] -= 4.0 * factor;
    _H.data[n1+1][n1+1] += 6.0 * factor;
    _H.data[n2-1][n2-1] += 6.0 * factor;
    _H.data[n1+1][n1+2] -= 4.0 * factor;
    _H.data[n2-1][n2-2] -= 4.0 * factor;
    _H.data[n1+1][n1+3] += factor;
    _H.data[n2-1][n2-3] += factor;
    for (i = n1+2; i <= n2-2; i++) {
      _H.data[i][i] += 6.0 * factor;
      _H.data[i][i-1] -= 4.0 * factor;
      _H.data[i][i+1] -= 4.0 * factor;
      _H.data[i][i-2] += factor;
      _H.data[i][i+2] += factor;
    }
  }
  return OE_SUCCESS;
}

/* Add a flatness constraint to the elements of the state vector
   between n1 and n2 inclusive, with a specified flatness factor. The
   matrix penalises gradients in x between n1 and n2.  */
int
oe_register_flatness_constraint(OEreal factor, int n1, int n2)
{
  int i;
  if (n1 >= n2 || n2 >= _H.n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Flatness constraint parameters out of range");
  }
  else if (n2 < n1+1) {
    /* Cannot apply a flatness constraint to fewer than 2 points*/
    return OE_SUCCESS;
  }
  is_smoothness_constraint = 1;
  /* Do the ends */
  _H.data[n1][n1] += factor;
  _H.data[n2][n2] += factor;
  _H.data[n1][n1+1] -= factor;
  _H.data[n2][n2-1] -= factor;
  for (i = n1+1; i <= n2-1; i++) {
    _H.data[i][i] += 2.0 * factor;
    _H.data[i][i-1] -= factor;
    _H.data[i][i+1] -= factor;
  }
  return OE_SUCCESS;
}

/* Prepare the matrices for the solver. The Hessian is calculated and
   returned in A_out, along with the right-hand side of the equation
   in rhs_out, such that the solver then finds x by solving the linear
   system of equations A_out*x=rhs_out. */
int
oe_prepare(OEvector *x, /* Current estimate of the state vector */
	   OEvector *y, /* Observations */
	   OEvector *y_model, /* Forward model estimate of observations */
	   OEmatrix *K, /* Jacobian */
	   OEvector *inv_s, /* Inverse of the error variance of each obs */
	   OEmatrix *A_out, /* Hessian (output) */
	   OEvector *rhs_out /* Right-hand side (output) */
	   )
{
  int m = K->m;
  int n = K->n;
  int h, i, j;

  /* Check sizes of matrices and vectors match */
  if (m != y->n || m != y_model->n || m != inv_s->n
      || n != A_out->m || n != A_out->n || n != rhs_out->n
      || n != x->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Matrix dimensions do not agree");
  }

  OE_CHECK(oe_reset_matrix(&_M, m, n));
  OE_CHECK(oe_reset_matrix(A_out, n, n));

  /* Initialise A_out = (Sa-1 + K' _M) with the Sa-1 term(s) */
  OE_CHECK(oe_zero_matrix(A_out));

  for (j = 0; j < OE_MAX_PRIOR_CONSTRAINTS; j++) {
    if (is_a_priori[j]) {
      if (n != _xa[j].n) {
	OE_THROW(OE_DIMENSION_ERROR,
		 "Size of a priori does not match state vector");
      }
      if (a_priori_is_diagonal[j]) {
	if (n != _inv_sa[j].n) {
	  OE_THROW(OE_DIMENSION_ERROR,
   "Size of a priori inverse error variance does not match state vector");
	}
	for (i = 0; i < n; i++) {
	  A_out->data[i][i] += _inv_sa[j].data[i];
	}
      }
      else {
	if (n != _inv_Sa[j].n) {
	  OE_THROW(OE_DIMENSION_ERROR,
   "Size of a priori inverse error covariance does not match state vector");
	}
	OE_CHECK(oe_add_matrix(&_inv_Sa[j], A_out));
      }
    }
  }

  /* Add smoothness matrix */
  if (is_smoothness_constraint) {
    if (n != _H.n) {
      OE_THROW(OE_DIMENSION_ERROR,
	       "Size of Twomey matrix does not match state vector");
    }
    OE_CHECK(oe_add_band_diagonal(&_H, 2, A_out));
  }

  /* Matrix multiplication: needs to be done efficiently as otherwise
     this can be the slowest part of the whole retrieval */
#ifndef OE_NO_SPARSE
  OE_CHECK(oe_sparse_KT_invs_K(K, inv_s, A_out, &_M));
#else
  OE_CHECK(oe_KT_invs_K(K, inv_s, A_out, &_M));
#endif

  /* Calculate rhs_out = _M'(y-y_model) - Sa-1(x-xa) */
  OE_CHECK(oe_zero_vector(rhs_out));
  for (j = 0; j < m; j++) {
    OEreal delta_y = y->data[j] - y_model->data[j];
    for (i = 0; i < n; i++) {
      rhs_out->data[i] += _M.data[j][i] * delta_y;
    }
  }

  /* Subtract the Sa-1(x-xa) term(s) */
  for (j = 0; j < OE_MAX_PRIOR_CONSTRAINTS; j++) {
    if (is_a_priori[j]) {
      if (a_priori_is_diagonal[j]) {
	for (i = 0; i < n; i++) {
	  rhs_out->data[i] -= _inv_sa[j].data[i] * (x->data[i]
						    - _xa[j].data[i]);
	}
      }
      else {
	for (i = 0; i < n; i++) {
	  OEreal delta_x_a = x->data[i] - _xa[j].data[i];
	  for (h = 0; h < n; h++) {
	    rhs_out->data[h] -= _inv_Sa[j].data[h][i] * delta_x_a;
	  }
	}
      }
    }
  }

  /* Subtract the Hx term */
  if (is_smoothness_constraint) {
    for (i = 0; i < n; i++) {
      int hmax = (i+2<n-1) ? i+2 : n-1;
      h = (i>2) ? (i-2) : 0;
      for ( ; h <= hmax; h++) {
	rhs_out->data[h] -= _H.data[h][i] * x->data[i];
      }
    }
  }    
  return OE_SUCCESS;
}

/* Calculate chi2 = SUM((y-y_model)^2/sigma(y)^2) to provide
   information as to when the iteration has converged. The input
   vectors are the observations (y), the forward model estimate of the
   observations (y_model) and the inverse of the error variance of the
   observations (inv_s). The single value of chi2 is returned in
   chi_squared_out. */
int
oe_chi_squared(OEvector *y, OEvector *y_model, OEvector *inv_s,
	       OEreal *chi_squared_out)
{
  int j, n = y->n;
  if (n != y_model->n || n != inv_s->n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Vector lengths do not match in chi squared test");
  }
  *chi_squared_out = 0.0;
  for (j = 0; j < n; j++) {
    *chi_squared_out += inv_s->data[j]
      * (y->data[j]-y_model->data[j]) * (y->data[j]-y_model->data[j]);
  }
  return OE_SUCCESS;
}

/* Subtract a priori constraints with index between jstart and jend
   (inclusive), from the Hessian matrix A. This is useful for
   calculating the error in the solution after the final iteration. */
int
oe_subtract_a_priori(OEmatrix *A, int jstart, int jend)
{
  int j;
  int n = A->n;

  /* Check arguments */
  if (A->m != n) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Input Hessian matrix is not square");
  }
  else if (jstart < 0 || jend >= OE_MAX_PRIOR_CONSTRAINTS) {
    OE_THROW(OE_DIMENSION_ERROR,
	     "Start and/or end constraint index out of range");
  }

  for (j = jstart; j <= jend; j++) {
    if (is_a_priori[j]) {
      if (a_priori_is_diagonal[j]) {
	int i;
	if (n != _inv_sa[j].n) {
	  OE_THROW(OE_DIMENSION_ERROR,
   "Size of a priori inverse error variance does not match input Hessian");
	}
	for (i = 0; i < n; i++) {
	  A->data[i][i] -= _inv_sa[j].data[i];
	}
      }
      else {
	if (n != _inv_Sa[j].n) {
	  OE_THROW(OE_DIMENSION_ERROR,
   "Size of a priori inverse error covariance does not match input Hessian");
	}
	OE_CHECK(oe_subtract_matrix(&_inv_Sa[j], A));
      }
    }
  }
  return OE_SUCCESS;
}
