/* readconfig.c -- Functions to read configuration information from a file

   Copyright (C) 2003 Robin Hogan <r.j.hogan@reading.ac.uk>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "readconfig.h"
#include "oe.h"

/* Move file pointer to the start of the next line, returning '\n' on
   success or EOF if the file ended first. */
static
int
__rc_skip_line(FILE *file)
{
  char c;
  do {
    c = fgetc(file);
  } while (c != '\n' && c != EOF);
  return c;
}

/* Move the file pointer over whitespace, returning the first
   non-whitespace character found, or '\n' if the line ended first, or
   EOF if the file ended first. */
static
int
__rc_skip_whitespace(FILE *file)
{
  int c;
  do {
    c = fgetc(file);
  } while (c <= ' ' && c != '\n' && c != EOF);
  return c;
}

/* Find param in data, returning an element of the rc_data list, or
   NULL if not present. Note that the search is case insensitive. */
static
rc_data *
__rc_find(rc_data *data, char *param)
{
  if (data->param) {
    if (strcasecmp(param, data->param) == 0) {
      return data;
    }
    else if (data->next) {
      return __rc_find(data->next, param);
    }
  }
  return NULL;
}

/* Add a param-value pair to an existing rc_data structure,
   overwriting an existing param with the same name (case
   insensitive). Note that param and value are the actual pointers
   added to the structure so they should not be freed after calling
   this function. Return 1 on success, 0 if memory allocation of the
   next element in the list failed. */
static
int
__rc_register(rc_data *data, char *param, char *value)
{
  char *c = param;
  int vector_dim_offset = 0;
  int m = 0, n = 0;

  /* Check if this is a vector and read in the dimensions */
  while (*c) {
    if (*c == '[') {
      vector_dim_offset = c - param;
      *c = '\0';
      break;
    }
    c++;
  }
  if (vector_dim_offset) {
    c++;
    if (*c) {
      char *newc;
      int val = strtol(c, &newc, 10);
      if (val > 0) {
	m = val;
      }
      if (newc && newc > c) {
	c = newc;
	while (*c && *c != '[') {
	  c++;
	}
	c++;
	if (*c) {
	  int val = strtol(c, &newc, 10);
	  if (val > 0) {
	    n = val;
	  }
	}
      }
    }
  }

  while (data->param) {
    if (strcasecmp(param, data->param) == 0) {
      if (data->value) {
	oe_free(data->value);
      }
      data->value = value;
      return 1;
    }
    data = data->next;
  }

  data->param = param;
  data->value = value;
  data->m = m;
  data->n = n;
  data->next = oe_malloc(sizeof(rc_data));
  if (!data->next) {
    return 0;
  }
  else {
    data->next->param = data->next->value = NULL;
    data->next->next = NULL;
    return 1;
  }
}

/* Read configuration information from file called file_name and
   return a pointer to the rc_data structure, or NULL if an error
   occurred.  If file_name is NULL then an empty rc_data structure is
   returned. If err_file is not NULL, errors messages will be written
   to err_file. */
rc_data *
rc_read(char *file_name, FILE *err_file)
{
  FILE *file;
  rc_data *data;
  char *memory_error = "Error allocating memory for configuration information\n";
  int c;

  if (!file_name) {
    /* Assign an empty data structure */
    data = (rc_data *) oe_malloc(sizeof(rc_data));
    if (!data) {
      if (err_file) {
	fprintf(err_file, memory_error);
      }
      return NULL;
    }
    data->param = NULL;
    data->value = NULL;
    data->next = NULL;
    data->m = data->n = 0;
    return data;
  }

  file = fopen(file_name, "r");

  if (!file) {
    if (err_file) {
      fprintf(err_file, "Error openning %s\n", file_name);
    }
    return NULL;
  }
  data = oe_malloc(sizeof(rc_data));
  data->param = NULL;
  data->value = NULL;
  data->next = NULL;
  data->m = data->n = 0;

  if (!data) {
    if (err_file) {
      fprintf(err_file, memory_error);
    }
    fclose(file);
    return NULL;
  }

  while (!feof(file)) {
    char *param = NULL, *value = NULL;
    int param_length = 0, value_length = 0;
    /* Skip whitespace */
    c = __rc_skip_whitespace(file);
    if (c == EOF) {
      break;
    }
    else if (c == '#') {
      __rc_skip_line(file);
      continue;
    }
    else if (c == '\n') {
      continue;
    }

    /* Read param name */
    while (c > ' ' && c != '#' && c != EOF) {
      param = oe_realloc(param, (++param_length)+1);
      if (!param) {
	if (err_file) {
	  fprintf(err_file, memory_error);
	}
	fclose(file);
	return NULL;
      }
      param[param_length-1] = c;
      c = fgetc(file);
    }
    ungetc(c, file);
    param[param_length] = '\0';

    /* Skip whitespace */
    c = __rc_skip_whitespace(file);
    if (c == '#') {
      __rc_skip_line(file);
    }
    else if (c != '\n') {
      /* Read value */
      if ((c == '\'') || (c == '"')) {
	int quote = c;
	c = fgetc(file);
	while (c != EOF && c != quote) {
	  value = oe_realloc(value, (++value_length)+1);
	  if (!value) {
	    if (err_file) {
	      fprintf(err_file, memory_error);
	    }
	    fclose(file);
	    return NULL;
	  }
	  value[value_length-1] = c;
	  value[value_length] = '\0';
	  c = fgetc(file);
	}
      }
      else if (c == '{') {
	c = fgetc(file);
	while (c != EOF && c != '}') {
	  if (c == '#') {
	    __rc_skip_line(file);
	  }
	  else {
	    value = oe_realloc(value, (++value_length)+1);
	    if (!value) {
	      if (err_file) {
		fprintf(err_file, memory_error);
	      }
	      fclose(file);
	      return NULL;
	    }
	    value[value_length-1] = c;
	    value[value_length] = '\0';
	  }
	  c = fgetc(file);
	}
      }
      else {
	while (c != EOF && c != '\n') {
	  if (c == '#') {
	    __rc_skip_whitespace(file);
	  }
	  else if (c != '\r') {
	    value = oe_realloc(value, (++value_length)+1);
	    if (!value) {
	      if (err_file) {
		fprintf(err_file, memory_error);
	      }
	      fclose(file);
	      return NULL;
	    }
	    value[value_length-1] = c;
	    value[value_length] = '\0';
	  }
	  c = fgetc(file);
	}
      }
    }
    /* Register result */
    if (!__rc_register(data, param, value)) {
      if (err_file) {
	fprintf(err_file, memory_error);
      }
      fclose(file);
      return NULL;
    }
  }
  fclose(file);
  return data;
}

/* Free an rc_data structure created with rc_read() - rc_data is a
   singly-linked list and its contents are recursively detelted. */
void
rc_clear(rc_data *data)
{
  if (data->param) {
    oe_free(data->param);
  }
  if (data->value) {
    oe_free(data->value);
  }
  if (data->next) {
    rc_clear(data->next);
  }
  oe_free(data);
}

/* Add a param-value pair to an existing rc_data structure,
   overwriting an existing param with the same name (case
   insensitive). Returns 1 on success and 0 on failure. */
int
rc_register(rc_data *data, char *param, char *value)
{
  char *newparam = oe_strdup(param);
  if (value) {
    char *newvalue = oe_strdup(value);
    return __rc_register(data, newparam, newvalue);
  }
  else {
    return __rc_register(data, newparam, NULL);
  }
}

/* Search the command-line arguments for param=value pairs and -param
   arguments and add them to the rc_data structure. */
int
rc_register_args(rc_data *data, int argc, char **argv)
{
  int i;
  for (i = 1; i < argc; i++) {
    /* Find an "=" sign in argument i */
    if (argv[i][0] == '-' && argv[i][1]) {
      char *param = oe_strdup(argv[i]+1);
      if (!param) {
	return 0;
      }
      if (!__rc_register(data, param, NULL)) {
	return 0;
      }
    }
    else {
      char *c = argv[i];
      while (*c != '\0') {
	if (*c == '=') {
	  /* Found one */
	  int param_length = c-argv[i];
	  char *value = oe_strdup(c+1);
	  char *param = oe_malloc(param_length+1);
	  if (!value || !param) {
	    return 0;
	  }
	  strncpy(param, argv[i], param_length);
	  param[param_length] = '\0';
	  if (!__rc_register(data, param, value)) {
	    return 0;
	  }
	}
	c++;
      }
    }
  }
  return 1;
}

/* Find the first config file on the command line, returning an index
   to the argv array, and 0 if none is found. Basically the first
   argument that contains no "=" sign and doesn't start with "-" is
   returned. However, after a "--" argument, a filename begining with
   "-" could in principle be returned. */
int
rc_get_file(int argc, char **argv)
{
  int i;
  int ignore_hyphen = 0;
  for (i = 1; i < argc; i++) {
    /* Find an "=" sign in argument i */
    char *c = argv[i];
    if (!ignore_hyphen && c[0] == '-') {
      if (strcmp(c, "--")) {
	ignore_hyphen = 1;
      }
      continue;
    }

    while (*c != '\0') {
      if (*c == '=') {
	break;
      }
      c++;
    }
    if (*c != '=') {
      return i;
    }
  }
  return 0;
}

/* Print contents of rc_data structure to file. */
void
rc_print(rc_data *data, FILE *file)
{
  char start_quote = '"', end_quote = '"';
  while (data->param) {
    fprintf(file, "%s", data->param);
    if (data->m > 0) {
      fprintf(file, "[%d]", data->m);
      if (data->n > 0) {
	fprintf(file, "[%d]", data->n);
      }
      start_quote = '{'; end_quote = '}';
    }
    if (data->value) {
      fprintf(file, " %c%s%c\n", start_quote, data->value, end_quote);
    }
    else {
      fprintf(file, " (no value)\n");
    }
    if (!(data = data->next)) {
      break;
    }   
  }
}

/* Return the contents of an rc_data structure as a string. Returns
   NULL on failure. Free with rc_free(). */
char *
rc_sprint(rc_data *data)
{
  int length = 0;
  char *out = NULL;
  while (data->param) {
    int param_length = strlen(data->param);
    out = oe_realloc(out, length+param_length+2);
    if (!out) {
      return NULL;
    }
    strcpy(out+length, data->param);
    length += (param_length+1);

    if (data->value) {
      int value_length = strlen(data->value);
      out[length-1] = ' ';
      out = oe_realloc(out, length+value_length+2);
      if (!out) {
	return NULL;
      }
      strcpy(out+length, data->value);
      length += (value_length+1);
    }

    out[length-1] = '\n';
    out[length] = '\0';

    if (!(data = data->next)) {
      break;
    }   
  }
  return out;
}

/* Return 1 if param exists in data, 0 otherwise. */
int
rc_exists(rc_data *data, char *param)
{
  return (__rc_find(data, param) != NULL);  
}

/* Interpret the value associated with param as a boolean, returning 1
   if true and 0 if false. Note that 0 will be returned if param is
   not present, or if it exists and is "0", "no" or "false" (or any
   case insensitive variants). Any other scenario will result in 1
   being returned */
int
rc_get_boolean(rc_data *data, char *param)
{
  data = __rc_find(data, param);
  if (!data) {
    return 0;
  }
  else if (!data->value) {
    return 1;
  }
  else if (strncasecmp(data->value, "false", 5) == 0
	   || strncasecmp(data->value, "no", 2) == 0) {
    return 0;
  }
  else {
    char *endptr = data->value;
    double val = strtod(data->value, &endptr);
    if (data->value == endptr || val != 0.0) {
      return 1;
    }
    else {
      return 0;
    }
  }
}

/* Interpret the value associated with param as an integer and return
   it. If successful *status is set to 1, but if either param is not
   present or the associated value cannot be interpreted as an
   integer, *status is set to 0. */
int
rc_get_int(rc_data *data, char *param, int *status)
{
  data = __rc_find(data, param);
  if (!data || !data->value) {
    *status = 0;
    return 0;
  }
  else {
    char *endptr;
    long val = strtol(data->value, &endptr, 10);
    if (data->value == endptr) {
      *status = 0;
      return 0;
    }
    else {
      *status = 1;
      return val;
    }
  }
}

/* If param exists in data, set *value to the value associated with
   param and return 1, otherwise leave *value untouched and return
   0. */
int
rc_assign_int(rc_data *data, char *param, int *value)
{
  int status;
  int val = rc_get_int(data, param, &status);
  if (status) {
    *value = val;
  }
  return status;
}

rc_real
rc_get_real(rc_data *data, char *param, int *status)
{
  data = __rc_find(data, param);
  if (!data || !data->value) {
    *status = 0;
    return 0;
  }
  else {
    char *endptr;
    double val = strtod(data->value, &endptr);
    if (data->value == endptr) {
      *status = 0;
      return 0;
    }
    else {
      *status = 1;
      return val;
    }
  }
}

/* As rc_assign_int() but with a real */
int
rc_assign_real(rc_data *data, char *param, rc_real *value)
{
  int status;
  rc_real val = rc_get_real(data, param, &status);
  if (status) {
    *value = val;
  }
  return status;
}

/* If param exists in data and can be interpretted as 1 or more
   reals, assign *value to a vector of reals, returning the number
   found. If fewer than min_length are present, the vector will be
   padded with the last valid real found up to min_length. In this
   case the number returned is the number of valid reals found, not
   min_length. If no reals are found or if there is an error
   allocating memory, *value is untouched and 0 is returned. */
int
rc_assign_real_vector(rc_data *data, char *param,
		      rc_real **value, int min_length)
{
  int length;
  rc_real *val = rc_get_real_vector(data, param, &length);
  if (!val) {
    return 0;
  }
  else if (length < min_length) {
    int i;
    val = oe_realloc(val, min_length*sizeof(rc_real));
    if (!val) {
      return 0;
    }
    for (i = length; i < min_length; i++) {
      val[i] = val[length-1];
    }
  }
  *value = val;
  return length;
}

/* As rc_assign_real_vector() except that if no reals are found then
   a vector of length "min_length" is returned containing
   default_value. This function always returns min_length, except if
   there was an error allocating memory in which case 0 is
   returned. */
int
rc_assign_real_vector_default(rc_data *data, char *param,
	      rc_real **value, int min_length, rc_real default_value)
{
  int length;
  rc_real *val = rc_get_real_vector(data, param, &length);
  if (!val) {
    int i;
    val = oe_malloc(min_length*sizeof(rc_real));
    if (!val) {
      return 0;
    }
    for (i = 0; i < min_length; i++) {
      val[i] = default_value;
    }
  }
  else if (length < min_length) {
    int i;
    val = oe_realloc(val, min_length*sizeof(rc_real));
    if (!val) {
      return 0;
    }
    for (i = length; i < min_length; i++) {
      val[i] = val[length-1];
    }
  }
  *value = val;
  return min_length;
}

/* Return the value associated with param as a string. NULL is
   returned on failure or memory allocation error. The string should
   be deallocated with rc_free().  */
char *
rc_get_string(rc_data *data, char *param)
{
  data = __rc_find(data, param);
  if (!data || !data->value) {
    return NULL;
  }
  else {
    char *value = oe_strdup(data->value);
    if (value) {
      /* Remove trailing whitespace */
      char *ch = value + strlen(value) - 1;
      while (*ch <= ' ') {
	*ch = '\0';
	ch--;
      }
    }
    return value;
  }
}

/* Free dynamically allocated data returned by rc_get_string(),
   rc_assign_string(), rc_assign_real_vector(),
   rc_assign_real_vector_default(), rc_get_int_vector() and
   rc_get_real_vector(). */
void
rc_free(void *string)
{
  if (string)
    oe_free(string);
}

/* Free dynamically allocated matrix returned by
   rc_get_real_matrix() */
void
rc_free_matrix(void **matrix)
{
  if (matrix) {
    if (matrix[0]) {
      oe_free(matrix[0]);
    }
    oe_free(matrix);
  }
}


/* If param exists in data, set *value to a pointer to a copy of the
   string associated with param and return 1, otherwise leave value
   untouched and return 0. Memory allocation error also results in 0
   being returned. The string should be freed with rc_free().  */
int
rc_assign_string(rc_data *data, char *param, char **value)
{
  char *str = rc_get_string(data, param);
  if (str) {
    if (*str) {
      /* A genuine non-empty string */
      *value = str;
      return 1;
    }
    else {
      /* An empty string */
      oe_free(str);
    }
  }
  return 0;
}

/* If param exists, interpret the associated value as integers and
   return a pointer to an int vector or NULL on failure. *length
   contains the number of integers assigned, 0 if none or memory
   allocation error. The vector should be freed with rc_free(). */
int *
rc_get_int_vector(rc_data *data, char *param, int *length)
{
  int *out = NULL;
  data = __rc_find(data, param);
  *length = 0;

  if (!data || !data->value) {
    return NULL;
  }
  else {
    char *endptr;
    char *c = data->value;
    while (*c) {
      long val = strtol(c, &endptr, 10);
      if (endptr != c) {
	out = oe_realloc(out, (++*length)*sizeof(int));
	if (!out) {
	  *length = 0;
	  return NULL;
	}
	out[*length-1] = val;
	c = endptr;
      }
      else {
	break;
      }
    }
    return out;
  }
}

/* As rc_get_real_vector() but with reals. */
rc_real *
rc_get_real_vector(rc_data *data, char *param, int *length)
{
  rc_real *out = NULL;
  data = __rc_find(data, param);
  *length = 0;

  if (!data || !data->value) {
    return NULL;
  }
  else {
    char *endptr;
    char *c = data->value;
    while (*c) {
      double val = strtod(c, &endptr);
      if (endptr != c) {
	out = oe_realloc(out, (++*length)*sizeof(rc_real));
	if (!out) {
	  *length = 0;
	  return NULL;
	}
	out[*length-1] = val;
	c = endptr;
      }
      else {
	break;
      }
    }
    return out;
  }
}

/* If param exists, interpret the associated value as reals in a 2D
   matrix and return a pointer to an array of pointers to each"row of
!  the matrix, or NULL on failure. *m and *n contain the number of
   rows and columns, respectively, 0 if none or memory allocation
   error. The matrix should be freed with rc_free_matrix(). */
rc_real **
rc_get_real_matrix(rc_data *data, char *param, int *m, int *n)
{
  int length, i;
  rc_real *val, **pval;
  data = __rc_find(data, param);

  if (!data) {
    return NULL;
  }
  val = rc_get_real_vector(data, param, &length);
  if (!val) {
    return NULL;
  }
  if (data->m * data->n != length) {
    rc_free(val);
    fprintf(stderr, "Error reading \"%s\": should have %d*%d=%d elements but found %d\n",
	    param, data->m, data->n, data->m*data->n, length);
    return NULL;
  }

  pval = oe_malloc(length*sizeof(rc_real*));
  if (!pval) {
    rc_free(val);
    return 0;
  }
  for (i = 0; i < data->m; i++) {
    pval[i] = val + i*data->n;
  }
  *m = data->m;
  *n = data->n;
  return pval;
}
