#include <iostream>
#include <fstream>
#include <cstring>
#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <map>

#include "rplot.h"
//#include "chil.h"
#include "mapplot.h"

#define UF_CAL 0
#define UF_PPI 1
#define UF_COPLANE 2
#define UF_RHI 3
#define UF_VERT 4
#define UF_DWELL 5
#define UF_MANUAL 6
#define UF_IDLE 7

extern "C" {
#include <netcdf.h>
}

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    cerr << "NetCDF error on line " << __LINE__ << ": " \
      << nc_strerror(ncstatus) << "\n"; \
    exit(-1); \
} 
#define report_error(a, s) if ((ncstatus = (a)) != NC_NOERR) { \
    cerr << "NetCDF error on line " << __LINE__ << ": " \
      << nc_strerror(ncstatus) << " (" << s << ")\n"; \
    exit(-1); \
} 

#define MAX_PLOTS 16
#define MAX_VAR_RANGES 256
#define NO_SIGNAL -999.0
#define NO_FLOAT_ATTRIBUTE -1.0e30
#define NO_FLOAT_ATTRIBUTE_TEST -0.9e30
#define PI_OVER_180 0.01745329
#define RADIUS_OF_THE_EARTH 6371.0

#define DEFAULT_RAY_DURATION 0.01

//#define DEBUG cerr << "Reached line " << __LINE__ << "\n"
#define DEBUG /* */

#define ERROR(msg,status) cerr << "Error: " << (msg) << "\n"; exit(status)
#define MINIMUM(a,b) (((a) > (b)) ? (b) : (a))
#define MAXIMUM(a,b) (((a) < (b)) ? (b) : (a))

enum FILE_FORMAT {
  FORMAT_POSTSCRIPT,
  FORMAT_PNG
};

static bool verbose = false, quiet = false;


/* Useful for certain applications */

typedef struct {
  float min;
  float max;
  float tick_interval;
  float tick_start;
  bool logarithmic;
  bool override;
} VarRange;

std::map<string,VarRange> var_range_list;

static
void
insert_var_range(char *var_name, char *var_range,
		 bool logarithmic=false, bool override=false)
{
  VarRange range;
  int n_assigned = sscanf(var_range, "%f:%f:%f:%f", 
			  &range.min, &range.tick_interval,
			  &range.max, &range.tick_start);
  if (n_assigned < 2) {
    cerr << "Error reading \"" << var_range
	 << "\" as a plot range for variable \"" << var_name << "\n";
    return;
  }
  else if (n_assigned == 2) {
    range.max = range.tick_interval;
    range.tick_interval = best_tick_spacing(range.min, range.max);
    range.tick_start = range.min;
  }
  else if (n_assigned == 3) {
    range.tick_start = range.min;
  }
  range.override = override;
  range.logarithmic = logarithmic;
  var_range_list[var_name] = range;
  //  var_range_list.insert(std::map<string,VarRange>::value_type(string(var_name),range));
  return;
};

static
void
insert_var_range(char *var_name, float min, float max, float tick_interval,
		 float tick_start, bool logarithmic=false, bool override=false)
{
  VarRange range = {min, max, tick_interval, tick_start, logarithmic, override};
  var_range_list[var_name] = range;
  //  var_range_list.insert(std::map<string,VarRange>::value_type(string(var_name),range));
  return;
}

static
void
insert_var_range(char *var_name, char *min, char *max,
		 bool logarithmic=false, bool override=false)
{
  VarRange range;
  if (!sscanf(min, "%f", &range.min)) {
    cerr << "Error reading \"" << min << "\" as a float\n";
    exit(-1);
  }
  if (!sscanf(max, "%f", &range.max)) {
    cerr << "Error reading \"" << max << "\" as a float\n";
    exit(-1);
  }
  range.tick_interval = best_tick_spacing(range.min, range.max);
  range.tick_start = range.min;
  range.logarithmic = logarithmic;
  range.override = override;
  var_range_list[var_name] = range;
  //  var_range_list.insert(std::map<string,VarRange>::value_type(string(var_name),range));
  return;
}

static
bool
get_var_range(char *var_name, VarRange *range)
{
  if (var_range_list.count(var_name)) {
    *range = var_range_list[var_name];
    return true;
  }
  *range = var_range_list["default"];
  return false;
}

static
void
print_var_range(char *var_name)
{
  VarRange range;
  get_var_range(var_name, &range);
  cerr << var_name << ": " << range.min << ", " << range.max
       << ", " << range.tick_interval << ", " << range.tick_start << "\n";
}

static
char *
request_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  if (nc_inq_atttype(ncid, varid, name, &type) != NC_NOERR) {
    return NULL;
  }
  if (type != NC_CHAR) {
    return NULL;
  }
  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    return NULL;
  }
  value = new char[length+1];
  if (!value) {
    cerr << "Out of memory\n";
    exit(2);
  }
  if (nc_get_att_text(ncid, varid, name, value) != NC_NOERR) {
    return NULL;
  }
  value[length] = '\0';
  return value;
}

static
char *
get_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  report_error(nc_inq_atttype(ncid, varid, name, &type), name);
  if (type != NC_CHAR) {
    cerr << "Error: attribute " << name << " is not of type char\n";
    exit(-1);
  }
  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  value = new char[length+1];
  if (!value) {
    cerr << "Out of memory\n";
    exit(2);
  }
  report_error(nc_get_att_text(ncid, varid, name, value), name);
  value[length] = '\0';
  return value;
}

static
short int
get_short_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  short int value;

  report_error(nc_inq_atttype(ncid, varid, name, &type), name);
  if (type != NC_SHORT && type != NC_INT) {
    cerr << "Error: attribute " << name << " is not of type short int or int\n";
    exit(-1);
  }
  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  if (length > 1) {
    cerr << "Error: attribute " << name << " has more than 1 element\n";
  }
  report_error(nc_get_att_short(ncid, varid, name, &value), name);
  return value;
}

static
int
request_short_attribute(int ncid, int varid, char *name, short *value)
{
  size_t length;
  nc_type type;

  if (nc_inq_atttype(ncid, varid, name, &type) == NC_NOERR
      && type == NC_SHORT
      && nc_inq_attlen(ncid, varid, name, &length) == NC_NOERR
      && length > 0) {
    report_error(nc_get_att_short(ncid, varid, name, value), name);
    return 0;
  }
  else {
    return 1;
  }
}

static
int
get_int_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  int value;

  report_error(nc_inq_atttype(ncid, varid, name, &type), name);
  if (type != NC_SHORT && type != NC_INT) {
    cerr << "Error: attribute " << name << " is not of type short int or int\n";
    exit(-1);
  }
  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  if (length > 1) {
    cerr << "Error: attribute " << name << " has more than 1 element\n";
  }
  report_error(nc_get_att_int(ncid, varid, name, &value), name);
  return value;
}

static
int
request_int_attribute(int ncid, int varid, char *name, int *value)
{
  size_t length;
  nc_type type;

  if (nc_inq_atttype(ncid, varid, name, &type) == NC_NOERR
      && type == NC_INT
      && nc_inq_attlen(ncid, varid, name, &length) == NC_NOERR
      && length > 0) {
    report_error(nc_get_att_int(ncid, varid, name, value), name);
    return 0;
  }
  else {
    return 1;
  }
}

static
float
get_float_attribute(int ncid, int varid, char *name)
{
  size_t length;
  float value;

  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    return NO_FLOAT_ATTRIBUTE;
  }
  if (length > 1) {
    cerr << "Error: " << name << " attribute has more than 1 element\n";
  }
  report_error(nc_get_att_float(ncid, varid, name, &value), name);
  return value;
}

static
float *
get_float_array_attribute(int ncid, int varid, char *name, size_t *length)
{
  float *value;

  if (nc_inq_attlen(ncid, varid, name, length) != NC_NOERR) {
    return NULL;
  }
  value = new float[*length];
  report_error(nc_get_att_float(ncid, varid, name, value), name);
  return value;
}

static
int
get_date(char *datestr, int day, int month, int year)
{
  static char *months[] = {"Jan","Feb","Mar","Apr","May","June",
			   "July","Aug","Sept","Oct","Nov","Dec"};
  if (month < 1 || month > 12) {
    return 1;
  }
  else if (day < 1) {
    sprintf(datestr, "%s %d", months[month-1], year);
  }
  else {
    sprintf(datestr, "%d %s %d", day, months[month-1], year);
  }
  return 0;
}

static
char *
strdup_tolower(const char *in)
{
  char *out = strdup(in);
  char *ch = out;
  if (out) {
    while (*ch) {
      *ch = tolower(*ch);
      ++ch;
    }
  }
  return out;
}

static
void
Usage(char **argv)
{
  cerr << "Usage: " << argv[0] << " [options] file1.nc [file2.nc ...]\n"
    "Options:\n"
    "    -hourrange hr1 hr2  set hour axis range\n"
    "    -plotmonth          plot an entire month\n"
    "     -dayrange d1 d2    plot multi-day images between days d1 and d2\n"
    "  -heightrange ht1 ht2  set height axis range\n"
    "       -rrange x1 x2    set horizontal axis range for RHIs\n"
    "     -varrange var rng  set colour scale range for variable \"var\" where \"rng\"\n"
    "                          is of the form \"min:max\", \"min:tick_interval:max\" or\n"
    "                          \"min:tick_interval:max:tick_start\"\n"
    "  -varlogrange var rng  as \"-varrange\" but using a log scale\n"
    "        -range f1 f2    set colour scale range for all plotted variables\n"
    "       -Zrange Z1 Z2    set colour scale range for Z (dBZ)\n"
    "       -vrange v1 v2    set colour scale range for velocity (m/s)\n"
    "-defaultvrange v1 v2    set velocity range but don't override folding velocity\n"
    "   -widthrange w1 w2    set colour scale range for spectral width (m/s)\n"
    "      -svrange sv1 sv2  set colour scale range for velocity standard deviation (m/s)\n"
    "   -PHIDPrange p1 p2    set the PHIDP colour scale range\n"
    "     -LDRrange l1 l2    set the LDR colour scale range\n"
    "    -rangename name     set the name of the range vector to use (e.g. height)\n"
    "       -nezcal f        calibrate radar by forcing nez to be f dBZ\n"
    "-plotlogmissing         on log plots, display missing data areas in the colour of\n"
    "                          the minimum value\n"
    "      -betacal f        calibration figure for lidar beta (dB)\n"
    "     -duration hrs      plot only specified number of hours - if negative,\n"
    "                          plot most recent data (overridden by `-hourrange')\n"
    "  -rayduration f        assumed duration of each ray in seconds (default "
       << DEFAULT_RAY_DURATION*3600.0 << ")\n"
    "         -size <x>x<y>  set size of axes in pixels\n"
    "      -barsize <x>x<y>  set size of colourbar in pixels\n"
    "     -defwidth x        width of the text definitions in classification data\n"
    "     -fontsize n        set pixel size of font\n"
    "       -system system   specify the name of the instrument\n"
    "        -field field    specify the full name of field being plotted\n"
    "        -units units    specify the units of the field\n"
    "   -fieldtitle          write the name of the field on each panel\n"
    "           -id id       string identifier used in the output filename\n"
    "     -onefield          plot only the first field from each file\n"
    "      -montage          produce only one plot containing all the variables\n"
    "     -onlyplot fields   only plot fields in comma-separated list (e.g. Zh,Zdr)\n"
    "       -stdout          send plot to standard output (implies -montage)\n"
    "          -mag f        set the magnification factor (default 1)\n"
    "      -maxsize d        scale down image if panel is larger than d pixels\n"
    "          -png          set image format to portable network graphics (default)\n"
    "     -ps, -eps          set image format to encapsulated postscript\n"
    "       -prefix dir      write the plots in specified directory\n"
    "      -palette file     read RGB colourmap from a file\n"
    "          -xfs server   specify the X font server (otherwise try unix/:-1,7100,7101,7102)\n" 
    "      -verbose          talk to me\n";
  //  CHILCreateFileNameUsage();
  exit(0);
}

int
main(int argc, char **argv)
{
  double minr = 0, maxr = 95;

  RFigure *figure[MAX_PLOTS];
  RAxes *axes[MAX_PLOTS];
  RPalette colormap;
  RFontLoader *font_loader;
  ofstream ofs[MAX_PLOTS];

  FILE_FORMAT format=FORMAT_PNG;
  char *xfs = "";
  char titlestr[512], paramstr[512];
  char *field_name[MAX_PLOTS];
  bool time_series[MAX_PLOTS];
  float last_value[MAX_PLOTS];
  /* these override values are for cloud data only, unless size_override == 1 */
  int width_override = 650, height_override = 230;
  int width = 500, height = 500;
  int bar_width = 16, bar_height = 160;
  int x_offset = 50, y_offset_default = 40;
  int y_offset[MAX_PLOTS];
  float x_mag = 1, y_mag = 1;
  int max_size = 0;
  int fig_width, fig_height;
  int definition_width = 100;

  int iarg, ifile, ivar;
  int max_vars = 0, nvars;
  int ncid, rangeid, rangedimid, timedimid, timeid, varid[MAX_PLOTS], nezid = 0;
  int elevid, azimid, freqid;
  size_t ngates, nrays, iray;
  float *range, *var = NULL, time, nexttime, nez, nez_override = 0.0;
  size_t start[2], count[2];
  int tmp;
  short int year, month, day;

  float minvalue[MAX_PLOTS], maxvalue[MAX_PLOTS];
  float intervalue[MAX_PLOTS], tickstartvalue[MAX_PLOTS];

  float missing_value[MAX_PLOTS], vfold[MAX_PLOTS];
  float defsize[MAX_PLOTS];
  float betacal = 0.0; /* In bels!!! */
  bool log_scale[MAX_PLOTS], using_nez[MAX_PLOTS];
  char *system = NULL, *system_override = NULL, *system_true = NULL;
  char *long_name = NULL, *long_name_override = NULL, *range_name_override = NULL;
  char *units = NULL, *units_override = NULL;
  char *suffix, *underscore1 = "", *underscore2 = "";
  char *locfile = NULL;
  char *system_code = "", *prefix = ".", *mapfile = NULL;
  //  char *filename_format = NULL;
  char *field_list = NULL;
  int font_size = 12;
  float starthour = 0, endhour = 24, minheight = 0, maxheight = 12;
  float startday = 1, endday = 32;
  float duration = 0, dhour, rayduration = DEFAULT_RAY_DURATION;
  float minx = -90, maxx = 90, miny = -90, maxy = 90;
  float angle = 0, dangle = 0, bestangle = 0;

  bool no_bar = false, one_field = false, one_plot = false;
  bool use_duration = false, override_betacal = false;
  bool nez_calibrate = false, setrrange = false;
  bool size_override = false, standard_output = false, field_title = false;
  bool plot_log_missing = false;
  bool multiday = false, firstday = true, lastday = false;

  /*
  float minval_override = -15.0, maxval_override = 15.0;
  float minZcloud = -50.0, maxZcloud = 10.0;
  float minZ = -20, maxZ = 40.0;
  float minv = -5.0, maxv = 5.0;
  // float minwidth = 0.0, maxwidth = 0.8; // This is for a linear scale
  float minwidth = -1.5, maxwidth = 0; // This is for a logarithmic scale
  
  // float minsv = 0.0, maxsv = 2.0; // This is for a linear scale
  float minsv = -3, maxsv = 0; // This is for a logarithmic scale
  float minZdr = -1, maxZdr = 4;
  float minLdr = -40, maxLdr = 0;
  float minphidp = -10, maxphidp = 40;
  */

  /* Set default colour ranges */
  insert_var_range("default", -15, 15, 10, -10);
  insert_var_range("Zcloud", -50, 20, 10, -50);
  insert_var_range("Zscanning", -20, 40, 10, -20);
  insert_var_range("v", -5, 5, 2, -4);
  insert_var_range("width", -1.5, 0, 0.5, -1.5, true);
  insert_var_range("width_h", -1.5, 0, 0.5, -1.5, true);
  insert_var_range("sigma_v", -3, 0,1, -3, true);
  insert_var_range("Zdr", -1, 4, 1, -1);
  insert_var_range("Ldr", -40, 0, 10, -40);
  insert_var_range("phi_dp", -10, 40, 10, -40);
  insert_var_range("uwind", -50, 50, 10, -50);
  insert_var_range("vwind", -50, 50, 10, -50);
  insert_var_range("wwind", -0.1, 0.1, 0.02, -0.1);
  insert_var_range("temperature", 200, 300, 10, 200);
  insert_var_range("pressure", 0, 110000, 10000, 0);
  insert_var_range("q", -6, -2, 1, -6, true);
  insert_var_range("ql", -7, -3, 1, -7, true);
  insert_var_range("qi", -7, -3, 1, -7, true);
  insert_var_range("cloud_fraction", 0, 1, 0.1, 0);
  insert_var_range("rh", 0, 1, 0.1, 0);
  insert_var_range("rhi", 0, 1, 0.1, 0);
  insert_var_range("rainrate", 0, 50, 5, 0);
  insert_var_range("lwp", -50, 500, 50, -50);
  /* Set default colour ranges for new Chilbolton format */
  insert_var_range("LDR_C", -30, -10, 5, -30);
  insert_var_range("SNR_HC", -10, 50, 10, -10);
  insert_var_range("SNR_VC", -10, 50, 10, -10);
  insert_var_range("SNR_XHC", -10, 50, 10, -10);
  insert_var_range("SPW_HC", -1.5, 0.5, 0.5, -1.5, true);
  insert_var_range("VEL_HC", -5, 2, 2, -4);
  insert_var_range("ZDR_C", -1, 4, 1, -1);
  insert_var_range("ZED_HC", -45, 35, 10, -40);
  insert_var_range("ZED_VC", -45, 35, 10, -40);
  insert_var_range("ZED_XHC", -45, 35, 10, -40);

  /* Parse switches */
  iarg=1;
  while (iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg], "-verbose") == 0) {
      verbose = true;
    }
    else if (strcmp(argv[iarg], "-onefield") == 0) {
      one_field = true;
    }
    else if (strcmp(argv[iarg], "-fieldtitle") == 0) {
      field_title = true;
    }
    else if (strcmp(argv[iarg], "-plotlogmissing") == 0) {
      plot_log_missing = true;
    }
    else if (strcmp(argv[iarg], "-plotmonth") == 0) {
      multiday = true;
    }
    else if (strcmp(argv[iarg], "-stdout") == 0) {
      one_plot = true;
      standard_output = true;
    }
    else if (strcmp(argv[iarg], "-montage") == 0) {
      one_plot = true;
    }
    else if (strcmp(argv[iarg], "-png") == 0) {
      format = FORMAT_PNG;
    }
    else if ((strcmp(argv[iarg],"-ps") == 0) 
	     || (strcmp(argv[iarg],"-eps") == 0)) {
      format = FORMAT_POSTSCRIPT;
    }
    else if (strcmp(argv[iarg],"-hourrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&starthour) < 1) {
	ERROR("Failed to read start hour as float", -1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&endhour) < 1) {
	ERROR("Failed to read end hour as float", -1);
      }
      use_duration = false;
    }
    else if (strcmp(argv[iarg],"-dayrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&startday) < 1) {
	ERROR("Failed to read start day as float", -1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&endday) < 1) {
	ERROR("Failed to read end day as float", -1);
      }
      use_duration = false;
      multiday = true;
    }
    else if (strcmp(argv[iarg],"-varrange") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read linear colour scale range", -1);
      }
      insert_var_range(argv[iarg+1], argv[iarg+2], false, true);
      iarg+=2;
    }
    else if (strcmp(argv[iarg],"-varlogrange") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read logarithmic colour scale range", -1);
      }
      insert_var_range(argv[iarg+1], argv[iarg+2], true, true);
      iarg+=2;
    }
    else if (strcmp(argv[iarg],"-range") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read colour scale range", -1);
      }
      insert_var_range("default", argv[iarg+1], argv[iarg+2], false, true);
      iarg+=2;
    }
    else if (strcmp(argv[iarg],"-rrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%lf",&minr) < 1) {
	ERROR("Failed to read minimum r as float", -1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%lf",&maxr) < 1) {
	ERROR("Failed to read maximum r as float", -1);
      }
      setrrange=true;
    }
    else if (strcmp(argv[iarg],"-Zrange") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read Z range", -1);
      }
      insert_var_range("Z", argv[iarg+1], argv[iarg+2], false, true);
      insert_var_range("Zh", argv[iarg+1], argv[iarg+2], false, true);
      iarg+=2;
    }
    else if (strcmp(argv[iarg],"-vrange") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read velocity range", -1);
      }
      insert_var_range("v", argv[iarg+1], argv[iarg+2], false, true);
      iarg+=2;
    }
    else if (strcmp(argv[iarg],"-defaultvrange") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read velocity range", -1);
      }
      insert_var_range("v", argv[iarg+1], argv[iarg+2], false, false);
      iarg+=2;
    }
    else if (strcmp(argv[iarg],"-widthrange") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read spectral width range", -1);
      }
      insert_var_range("width", argv[iarg+1], argv[iarg+2], false, true);
      insert_var_range("width_h", argv[iarg+1], argv[iarg+2], false, true);
      iarg+=2;
    }
    else if (strcmp(argv[iarg],"-svrange") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read velocity standard deviation range", -1);
      }
      insert_var_range("sigma_v", argv[iarg+1], argv[iarg+2], false, true);
      iarg+=2;
    }
    else if (strcmp(argv[iarg],"-heightrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minheight) < 1) {
	ERROR("Failed to read minimum height as float", -1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxheight) < 1) {
	ERROR("Failed to read maximum height as float", -1);
      }
    }
    else if (strcmp(argv[iarg],"-PHIDPrange") == 0) {
      if (argc == iarg+1) {
	ERROR("Failed to read PHIDP range",1);
      }
      insert_var_range("phi_dp", argv[iarg+1], argv[iarg+2], false, true);
      iarg+=2;
    }
    else if (strcmp(argv[iarg],"-LDRrange") == 0) {
      if (argc == iarg+1) {
	ERROR("Failed to read LDR range",1);
      }
      insert_var_range("Ldr", argv[iarg+1], argv[iarg+2], false, true);
      iarg+=2;
    }
    else if (strcmp(argv[iarg],"-mag") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&x_mag) < 1) {
	ERROR("Failed to read magnification value",1);
      }
      if (x_mag < 0.01) {
	x_mag = 0.01;
	cerr << "Using magnification factor of " << x_mag << "\n";
      }	
      else if (x_mag > 10) {
	x_mag = 10.0;
	cerr << "Using magnification factor of " << x_mag << "\n";
      }
      y_mag = x_mag;
    }
    else if (strcmp(argv[iarg],"-maxsize") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%d",&max_size) < 1) {
	ERROR("Failed to read maximum panel size", 1);
      }
    }
    else if (strcmp(argv[iarg], "-fontsize") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read font size",1);
      }
      font_size = atoi(argv[++iarg]);
      if (font_size < 1) {
	ERROR("Font size must be a positive integer",1);
      }
    }
    else if (strcmp(argv[iarg],"-duration") == 0) {
      if (argc > iarg+1) {
	duration = atof(argv[++iarg]);
	if (duration == 0.0) {
	  use_duration = false;
	}
	else {
	  use_duration = true;
	}
      }
      else {
	ERROR("Duration not specified", -1);
      }
    }
    else if (strcmp(argv[iarg],"-nezcal") == 0) {
      if (argc > iarg+1) {
	nez_override = atof(argv[++iarg]);
	nez_calibrate = true;
      }
      else {
	ERROR("Noise-equivalent reflectivity for calibration not specified", -1);
      }
    }
    else if (strcmp(argv[iarg],"-betacal") == 0) {
      if (argc > iarg+1) {
	betacal = 0.1 * atof(argv[++iarg]);
	override_betacal = true;
      }
      else {
	ERROR("Beta calibration figure not specified", -1);
      }
    }
    else if (strcmp(argv[iarg],"-rayduration") == 0) {
      if (argc > iarg+1) {
	rayduration = atof(argv[++iarg])/3600.0;
	if (rayduration <= 0.0) {
	  cerr << "Warning: ray duration less than or equal to 0; using "
	       << DEFAULT_RAY_DURATION*3600.0 << " seconds";
	  rayduration = DEFAULT_RAY_DURATION;
	}
      }
      else {
	ERROR("Ray duration not specified", -1);
      }
    }
    else if (strcmp(argv[iarg], "-size") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%dx%d",
				   &width_override, &height_override) < 2) {
	ERROR("Failed to read size: should be in the form "
	  "<width>x<height>", -1);
      }
      size_override = true;
    }
    else if (strcmp(argv[iarg], "-barsize") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%dx%d",
				   &bar_width,&bar_height) < 2) {
	ERROR("Failed to read bar size: should be in the form "
	      "<width>x<height>", -1);
      }
      if (bar_width < 1 || bar_height < 1) {
	no_bar = true;
      }
    }
    else if (strcmp(argv[iarg], "-defwidth") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read font size", -1);
      }
      definition_width = atoi(argv[++iarg]);
      if (font_size < 1) {
	ERROR("Definition width must be a positive integer", -1);
      }
    }
    else if (strcmp(argv[iarg], "-fontsize") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read font size", -1);
      }
      font_size = atoi(argv[++iarg]);
      if (font_size < 1) {
	ERROR("Font size must be a positive integer", -1);
      }
    }
    else if (strcmp(argv[iarg], "-units") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read units", -1);
      }
      units_override = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-system") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read system name", -1);
      }
      system_override = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-field") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read field name", -1);
      }
      long_name_override = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-onlyplot") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read field names", -1);
      }
      field_list = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-rangename") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read name of range vector", -1);
      }
      range_name_override = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-id") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read system ID", -1);
      }
      system_code = argv[++iarg];
      underscore1 = "_";
    }
    else if (strcmp(argv[iarg],"-prefix") == 0) {
      if (argc > iarg+1) {
	prefix = argv[++iarg];
      }
      else {
	ERROR("Prefix not specified", -1);
      }
    }
    /*
    else if (strcmp(argv[iarg],"-filename") == 0) {
      if (argc > iarg+1) {
	filename_format = argv[++iarg];
      }
      else {
	ERROR("Filename not specified", -1);
      }
    }
    */
    else if (strcmp(argv[iarg],"-coastline") == 0) {
      if (argc > iarg+1) {
	mapfile = argv[++iarg];
      }
      else {
	ERROR("Coastline file not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-locations") == 0) {
      if (argc > iarg+1) {
	locfile = argv[++iarg];
      }
      else {
	ERROR("Location file not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-palette") == 0) {
      if (argc > iarg+1) {
	mapfile = argv[++iarg];
      }
      else {
	ERROR("Map file not specified", -1);
      }
    }
    else if (strcmp(argv[iarg],"-xfs") == 0) {
      if (argc > iarg+1) {
	xfs = argv[++iarg];
      }
      else {
	ERROR("X font server not specified", -1);
      }
    }
    else {
      cerr << argv[iarg] << " not understood\n";
      Usage(argv);
    }
    ++iarg;
  }
  if (iarg > argc-1) {
    Usage(argv);
  }

  if (format == FORMAT_POSTSCRIPT) {
    suffix = "eps";
  }
  else {
    suffix = "png";
  }    

  try {

    /* Contact font server */
    font_loader = new RFontLoader(xfs, format == FORMAT_PNG);
    font_loader->verbose(verbose);

    if (mapfile) {
      if (colormap.load(mapfile)) {
	if (!quiet) {
	  cerr << "Warning: " << mapfile << " not found\n";
	}
	create_jet_palette(&colormap,64);
      }
    }
    else {
      create_jet_palette(&colormap,64);
    }

    /* Loop through all file names on the command line */
    for (ifile = iarg; ifile < argc; ++ifile) {
      size_t ind;
      char *scan_type = "Fixed", *scan_type3, *scan_type3_lower;
      char datestr[128];
      float frequency, start_time, end_time, fixduration;
      float x_scale = 1, y_scale = 1;
      unsigned short int file = 0, raster = 0, scan = 0;
      unsigned short int hour, minute, second;
      int i;
      unsigned char sweep_mode;
      bool cloud_data;
      bool time_in_s;
      int is_definition = 0;

      if (ifile == argc-1) {
	lastday = true;
      }

      /* Get dimension sizes */
      cerr << argv[ifile] << endl;
      handle_error(nc_open(argv[ifile], NC_NOWRITE, &ncid));
      handle_error(nc_inq_dimid(ncid, "time", &timedimid));
      handle_error(nc_inq_dimlen(ncid, timedimid, &nrays));

      if (range_name_override) {
	if(nc_inq_dimid(ncid, range_name_override, &rangedimid) != NC_NOERR) {
	  cerr << "Error reading " << range_name_override << endl;
	  exit(-1);
	}
	if (strcmp(range_name_override, "level") == 0) {
	  /* We have model data */
	  handle_error(nc_inq_varid(ncid, "height", &rangeid));
	}
	else {
	  handle_error(nc_inq_varid(ncid, range_name_override, &rangeid));
	}
      }
      else {
	if (nc_inq_dimid(ncid, "range", &rangedimid) != NC_NOERR) {
	  handle_error(nc_inq_dimid(ncid, "height", &rangedimid));
	  handle_error(nc_inq_varid(ncid, "height", &rangeid));
	}
	else {
	  handle_error(nc_inq_varid(ncid, "range", &rangeid));
	}
      }

      /* Get the ranges */
      handle_error(nc_inq_dimlen(ncid, rangedimid, &ngates));
      handle_error(nc_inq_varndims(ncid, rangeid, &tmp));
      if (tmp == 1) {
	start[0] = 0;
	count[0] = ngates;
      }
      else if (tmp == 2) {
	/* Range is a 2D field - we assume it doesn't change very much
           (true for model heights) and just take the first column. */
	start[0] = 0;
	start[1] = 0;
	count[0] = 1;
	count[1] = ngates;
      }
      else {
	ERROR("Range/height variable not 1 or 2 dimensions", -1);
      }
      range = new float[ngates];
      handle_error(nc_get_vara_float(ncid, rangeid, start, count, range));
      
      if (range[0] < -200) {
	/* Bad RACMO data? Try next column */
	start[0] = 1;
	handle_error(nc_get_vara_float(ncid, rangeid, start, count, range));
      }      

      char *rangeunits = get_char_attribute(ncid, rangeid, "units");
      if (strcmp(rangeunits, "m") == 0) {
	for (i = 0; i < ngates; i++) {
	  range[i] = range[i] * 0.001;
	}
      }

      /* Find time variable and get the start and end time */
      handle_error(nc_inq_varid(ncid, "time", &timeid));
      char *time_units = get_char_attribute(ncid, timeid, "units");

      ind = 0;
      handle_error(nc_get_var1_float(ncid, timeid, &ind, &start_time));
      ind = nrays-1;
      handle_error(nc_get_var1_float(ncid, timeid, &ind, &end_time));
      if (strncmp(time_units, "s", 1) == 0) {
	time_in_s = true;
	hour = (unsigned short int) start_time / 3600;
	minute = (unsigned short int) (start_time - hour * 60);
	second = (unsigned short int) (start_time - hour * 3600 - minute * 60);      

	fixduration = (end_time - start_time) / 3600;
      }
      else {
	time_in_s = false;
	hour = (unsigned short int) start_time;
	minute = (unsigned short int) (60.0 * (start_time - (float) hour));
	second = (unsigned short int) (60.0 * (60.0 * (start_time - (float) hour)
					       - (float) minute));
	fixduration = end_time - start_time;
      }

      if (fixduration <= 0) {
	fixduration += 24.0;
      }
      fixduration = ceil(fixduration * 60.0);

      /* Get the radar frequency */
      ind = 0;
      if (nc_inq_varid(ncid, "frequency", &freqid) == NC_NOERR) {
	handle_error(nc_get_var1_float(ncid, freqid, &ind, &frequency));
      }
      else {
	frequency = 0;
      }

      /* Set the title of the plot (stupidly called "system") and set
	 system_true */
      system_true = request_char_attribute(ncid, NC_GLOBAL, "system");
      if (!system_true) {
	system_true = "";
      }
      if (system_override) {
	system = system_override;
      }
      else {
	system = request_char_attribute(ncid, NC_GLOBAL, "title");
	if (!system) {
	  system = new char[128];
	  sprintf(system, "%d-GHz %s radar", (int) (frequency + 0.49), system_true);
	}
      }

      /* Get the date */
      if (request_short_attribute(ncid, NC_GLOBAL, "day", &day) == 0) {
	month = get_short_attribute(ncid, NC_GLOBAL, "month");
	year = get_short_attribute(ncid, NC_GLOBAL, "year");
      }
      else if (request_int_attribute(ncid, NC_GLOBAL, "day", &tmp) == 0) {
	day = (short) tmp;
	month = get_int_attribute(ncid, NC_GLOBAL, "month");
	year = get_int_attribute(ncid, NC_GLOBAL, "year");
      }
      else {
	if (sscanf(time_units+12, "%hu-%hu-%hu", &year, &month, &day) < 3) {
	  cerr << "Error interpretting time units \"" << time_units
	       << "\" in terms of YYYY-MM-DD" << endl;
	}
      }
      if (!multiday) {
	get_date(datestr, day, month, year);
      }
      else {
	get_date(datestr, 0, month, year);
      }

      /* Is it cloud data? */
      cloud_data = true;
      if (nc_inq_varid(ncid, "azim", &azimid) == NC_NOERR
	  && nc_inq_varid(ncid, "elev", &elevid) == NC_NOERR) {
	int nelevdims = 0, nazimdims = 0;
	if (nc_inq_varndims(ncid, azimid, &nazimdims) == NC_NOERR
	    && nc_inq_varndims(ncid, elevid, &nelevdims) == NC_NOERR
	    && nazimdims == 1 && nelevdims == 1) {
	  cloud_data = false;
	}
      }

      if (!cloud_data) {
	size_t tmp;
	/* Get the scan numbers and the radar frequency */
	/* Accept tape or file attributes */
	if (nc_inq_attlen(ncid, NC_GLOBAL, "tape", &tmp) == NC_NOERR) {
	  file = get_short_attribute(ncid, NC_GLOBAL, "tape");
	}
	else {
	  file = get_short_attribute(ncid, NC_GLOBAL, "file");
	}
	raster = get_short_attribute(ncid, NC_GLOBAL, "raster");
	scan = get_short_attribute(ncid, NC_GLOBAL, "scan");
      }

      /* Get the scan type */
      sweep_mode = UF_VERT;
      scan_type3 = "FIX";
      scan_type3_lower = "fix";
      if (!cloud_data) {
	scan_type = get_char_attribute(ncid, NC_GLOBAL, "scantype");
	if (strcmp(scan_type, "RHI") == 0) {
	  sweep_mode = UF_RHI;
	  scan_type3 = "RHI";
	  scan_type3_lower = "rhi";
	}
	else if (strcmp(scan_type, "Slant-plane") == 0) {
	  sweep_mode = UF_RHI;
	  scan_type3 = "S-P";
	  scan_type3_lower = "s-p";
	}
	else if (strcmp(scan_type, "PPI") == 0) {
	  sweep_mode = UF_PPI;
	  scan_type3 = "PPI";
	  scan_type3_lower = "ppi";
	}
	else if (strcmp(scan_type, "Calibration") == 0) {
	  scan_type3 = "CAL";
	  scan_type3_lower = "cal";
	}
	else if (strcmp(scan_type, "Manual") == 0) {
	  scan_type3 = "MAN";
	  scan_type3_lower = "man";
	}
      }

      if (firstday || !multiday) {
	firstday = false;
      if (cloud_data) {
	/*** CLOUD DATA ***/
	width = width_override;
	height = height_override;

	/* Set the hour range if necessary */
	if (use_duration) {
	  start[0] = 0;
	  handle_error(nc_get_var1_float(ncid, timeid, start, &starthour));


	if (time_in_s) {
	    starthour = starthour / 3600;
	}

	  if (starthour > 0.0 && starthour < duration*0.1) {
	    starthour = 0.0;
	  }
	  if (duration > 0.0) {
	    endhour = starthour + duration;
	  }
	  else {
	    start[0] = nrays-1;
	    handle_error(nc_get_var1_float(ncid, timeid, start, &endhour));
	    if (time_in_s) {
	     endhour = endhour / 3600;
	    }
	    if (endhour < 24.0 && endhour-24.0 > duration*0.1) {
	      endhour = 24.0;
	    }
	    if (endhour-starthour < (-duration)) {
	      endhour = starthour - duration;
	    }
	    else if ((starthour = endhour + duration) < 0) {
	      starthour = 0;
	      endhour = -duration;
	    }	   
	  }
	}
	if (!multiday) {
	  minx = starthour;
	  maxx = endhour;
	}
	else {
	  minx = startday;
	  maxx = endday;
	}
	miny = minheight;
	maxy = maxheight;
      }
      else {
	/*** SCANNING DATA ***/
	size_t index = 0;
	float azim1, azim2;
	float elev1, elev2;
	float ang;

	handle_error(nc_get_var1_float(ncid, azimid, &index, &azim1));
	index = nrays-1;
	handle_error(nc_get_var1_float(ncid, azimid, &index, &azim2));

	index = 0;
	handle_error(nc_get_var1_float(ncid, elevid, &index, &elev1));
	index = nrays-1;
	handle_error(nc_get_var1_float(ncid, elevid, &index, &elev2));

	angle = -1000;

	switch (sweep_mode) {
	case UF_RHI: case UF_COPLANE:
	  sprintf(titlestr,"<B>%s</B>", system);
	  sprintf(paramstr,"%s %02hu:%02hu:%02hu %s %04hu/%03hd/%02hu Azim %5.1f<SUP>o",
		  datestr, hour, minute, second,
		  scan_type3, file, raster, scan, 0.5*(azim1+azim2));

	  y_scale = 4;
	  if (strcmp(system_true, "Singapore") == 0) {
	    x_scale = 4;
	  }
	  miny = minheight;
	  maxy = maxheight;
	  if (setrrange) {
	    minx = minr;
	    maxx = maxr;
	  }
	  else {
	    minx = range[0] * cos(MAXIMUM(elev1, elev2) * PI_OVER_180);
	    maxx = MAXIMUM(range[ngates-1], 90.0/x_scale);
	    if (maxx < minx + 90.0/x_scale) {
	      if (maxx < 90.0/x_scale) {
		minx = 0;
		maxx = 90.0/x_scale;
	      }
	      else {
		minx = maxx - 90.0/x_scale;
	      }
	    }
	  }
	  break;
	case UF_PPI:

	  if (azim1 > azim2) {
	    float tmp_float = azim2;
	    azim2 = azim1;
	    azim1 = tmp_float;	    
	  }

	  minx = miny = 100;
	  maxx = maxy = -100;

	  for (ang = azim1; ang < azim2 + 1.0; ang += 1.0) {
	    /* Note: the +90.0 is because `angle' is CLOCKWISE from DUE WEST */
	    float ang_rad = (ang+90.0) * PI_OVER_180;
	    minx = MINIMUM(minx, range[ngates-1]*(-cos(ang_rad)));
	    minx = MINIMUM(minx, range[0]*(-cos(ang_rad)));
	    maxx = MAXIMUM(maxx, range[ngates-1]*(-cos(ang_rad)));
	    maxx = MAXIMUM(maxx, range[0]*(-cos(ang_rad)));
	    miny = MINIMUM(miny, range[ngates-1]*(sin(ang_rad)));
	    miny = MINIMUM(miny, range[0]*(sin(ang_rad)));
	    maxy = MAXIMUM(maxy, range[ngates-1]*(sin(ang_rad)));
	    maxy = MAXIMUM(maxy, range[0]*(sin(ang_rad)));
	  }
	  if (minx >= maxx-90) {
	    minx -= 0.5 * (90.0-maxx+minx);
	    maxx += 0.5 * (90.0-maxx+minx);
	  }
	  if (miny >= maxy) {
	    miny = -90.0; maxy = 90.0;
	  }
	  sprintf(titlestr,"<B>%s</B>",system);
	  sprintf(paramstr,"%s %02hu:%02hu:%02hu %s %04hu/%03hd/%02hu Elev %4.1f<SUP>o",
		  datestr, hour, minute, second,
		  scan_type3, file, raster, scan, 0.5*(elev1+elev2));
	  break;
	default:
	  y_scale = 4;
	  minx = 0;
	  maxx = fixduration;
	  miny = minheight;
	  maxy = maxheight;
	  x_scale = 100.0/fixduration;
	  sprintf(titlestr,"<B>%s</B>", system);
	  sprintf(paramstr,"%s %02hu:%02hu:%02hu %s %hu/%hd/%hu",
		  datestr, hour, minute, second,
		  scan_type3, file, raster, scan);
	}
	if (size_override) {
	  width = width_override;
	  height = height_override;
	}
	else {
	  width = (int) ((maxx-minx) * x_scale * x_mag * 5);
	  height = (int) ((maxy-miny) * y_scale * y_mag * 5);
	}

	minx = floor(minx);
	miny = floor(miny);
	maxx = ceil(maxx);
	maxy = ceil(maxy);

      }

      /*** ALL DATA ***/
      /* Check width & height do not exceed max_size */
      if (max_size > 0) {
	if (width > height) {
	  if (width > max_size) {
	    int tmp_width = width;
	    width = max_size;
	    height = (max_size * height) / tmp_width;
	  }
	}
	else {
	  if (height > max_size) {
	    int tmp_height = height;
	    height = max_size;
	    width = (max_size * width) / tmp_height;
	  }
	}
      }

      /* Find 2D variables to plot */
      handle_error(nc_inq_nvars(ncid, &nvars));
      max_vars = 0;
      for (ivar = 0; ivar < nvars; ++ivar) {
	char name[NC_MAX_NAME];
	nc_type xtype;
	int ndims, natts;
	int dimids[NC_MAX_VAR_DIMS];

	nc_inq_var(ncid, ivar, name, &xtype, &ndims, dimids, &natts);
	int namelen = strlen(name);
   	/* Is it 2D etc? */
	if (ndims == 2 && dimids[0] == timedimid && dimids[1] == rangedimid) {
	  if (field_list) {
	    /* Is it in our list of fields to plot? */
	    char *c1 = field_list;
	    bool found = false;

	    while ((c1 = strstr(c1, name))) {
	      char *c2 = c1+namelen;
	      if ((c1 == field_list || c1[-1] == ',')
		  && (*c2 == '\0' || *c2 == ',')) {
		found = true;
		break;
	      }
	      c1 = c2;
	    }
	    if (!found) {
	      /* Skip to next variable */
	      continue;
	    }
	  }
	  else if ((namelen > 4 && strcmp(name+namelen-4,"_raw") == 0)
		   || (namelen > 5 && strcmp(name+namelen-5,"_bias") == 0)) {
	    /* Skip over variables called *_raw or *_bias */
	    continue;
	  }
	  if (nc_inq_atttype(ncid, ivar, "units", &xtype) != NC_NOERR
	      || xtype != NC_CHAR) {
	    size_t n_plot_range = 0;
	    float *plot_range;
	    plot_range = get_float_array_attribute(ncid, ivar, "plot_range",
						   &n_plot_range);
	    if (!plot_range || n_plot_range < 2) {
	      /* Skip over variables that have no units and no
		 plot_range attributes */
	      /* Variables with no units but a plot_range are
		 interpretted as classification data */
	      continue;
	    }
	    else {
	      delete[] plot_range;
	      if (nc_inq_atttype(ncid, ivar, "definition", &xtype) == NC_NOERR
		  && xtype == NC_CHAR) {
		is_definition = 1;
	      }
	    }
	  }
	  varid[max_vars] = ivar;
	  field_name[max_vars] = strdup(name);
	  time_series[max_vars] = false;
	  ++max_vars;
	  if (max_vars >= MAX_PLOTS) {
	    break;
	  }
	}

	if (max_vars == 1 && one_field) {
	  break;
	}
      }
      if (field_list && !(max_vars == 1 && one_field)) {
	/* Look for 1D requested fields */
	for (ivar = 0; ivar < nvars; ++ivar) {
	  char name[NC_MAX_NAME];
	  nc_type xtype;
	  int ndims, natts;
	  int dimids[NC_MAX_VAR_DIMS];

	  nc_inq_var(ncid, ivar, name, &xtype, &ndims, dimids, &natts);
	  int namelen = strlen(name);
	  /* Is it 1D etc? */
	  if (ndims == 1 && dimids[0] == timedimid) {
	    /* Time series - might be one of the requested fields */
	    char *c1 = field_list;
	    bool found = false;

	    while ((c1 = strstr(c1, name))) {
	      char *c2 = c1+namelen;
	      if ((c1 == field_list || c1[-1] == ',')
		  && (*c2 == '\0' || *c2 == ',')) {
		found = true;
		break;
	      }
	      c1 = c2;
	    }
	    if (!found) {
	      /* Skip to next variable */
	      continue;
	    }
	    varid[max_vars] = ivar;
	    field_name[max_vars] = strdup(name);
	    time_series[max_vars] = true;
	    last_value[max_vars] = 0.0;
	    ++max_vars;
	    if (max_vars >= MAX_PLOTS) {
	      break;
	    }
	    else if (max_vars == 1 && one_field) {
	      break;
	    }
	  }
	}
      }
      if (!max_vars) {
	/* No fields to plot! */
	delete[] range;
	continue;
      }

      /* Work out the size of the figure */
      if (no_bar) {
	fig_width = width + x_offset + 20 + is_definition*definition_width;
      }
      else {
	fig_width = width + x_offset + bar_width + 64 + is_definition*definition_width;
      }
      if (one_plot) {
	fig_height = (height+y_offset_default)*max_vars + 40;
      }
      else {
	fig_height = height + y_offset_default + 40;
      }

      /* Set up each variable for plotting */
      for (i = 0; i < max_vars; i++) {
	RAxes *cbar;
	char tmpstr[256];
	float scale_offset;
	char *field;
	char name[NC_MAX_NAME];
	nc_type xtype;
	int ndims, natts;
	int dimids[NC_MAX_VAR_DIMS];
	int ivar = varid[i];
	size_t n_plot_range;
	float *plot_range;
	char *plot_scale;
	bool non_dimensional = false;

	nc_inq_var(ncid, ivar, name, &xtype, &ndims, dimids, &natts);

	if (units_override) {
	  units = units_override;
	}
	else {
	  units = request_char_attribute(ncid, ivar, "units_html");
	  if (!units) {
	    units = request_char_attribute(ncid, ivar, "units");
	    if (!units || (units[0] == '1' && units[1] == '\0')) {
	      non_dimensional = true;
	    }
	  }
	}

	if (long_name_override) {
	  long_name = long_name_override;
	}
	else {
	  long_name = get_char_attribute(ncid, ivar, "long_name");
	}

	missing_value[i] = get_float_attribute(ncid, ivar, "missing_value");

	if (one_field || one_plot) {
	  field = "";
	  underscore2 = "";
	}
	else {
	  field = name;
	  underscore2 = "_";
	}

	if (!standard_output) {
	  if (!one_plot || i == 0) {
//	      /* Use the user-specified format string to create a filename */
	    /*
	    if (filename_format) {
	      if (!CHILCreateFileName(tmpstr, filename_format, prefix,
				      file, raster, scan,
				      year, month, day,
				      hour, minute, second,
				      system_true, scan_type)) {
		cerr << "Error interpretting filename format \""
		     << filename_format << "\": "
		     << tmpstr << "\n";
		break;
	      }
	    */
	    //	    }
	    //	    else 
	    if (cloud_data) {
	      if (!multiday) {
		sprintf(tmpstr, "%s/%04hd%02hd%02hd%s%s%s%s.%s",
			prefix, year, month, day, underscore1,
			system_code, underscore2, field, suffix);
	      }
	      else {
		sprintf(tmpstr, "%s/%04hd%02hd%s%s%s%s.%s",
			prefix, year, month, underscore1,
			system_code, underscore2, field, suffix);
	      }
	    }
	    else {
	      //	      char *system_lower = strdup_tolower(system_true);
	      sprintf(tmpstr, "%s/%04hdr%03hds%02hd_%02hd%02hd%02hd_%s_%s%s%s%s%s.%s",
		      prefix, file, raster, scan, hour, minute, second,
		      scan_type3, system_true, underscore1, system_code,
		      underscore2, field, suffix);
	    }
	    ofs[i].open(tmpstr);

	    if (!ofs[i]) {
	      cerr << "Failed to write " << tmpstr << "\n";
	      exit(-1);
	    }

	    if (!quiet) {
	      if (!i) {
		cerr << "Writing";
	      }
	      cerr << " " << tmpstr ;
	    }

	    /* Initialise plot */
	    if (format == FORMAT_POSTSCRIPT) {
	      figure[i] = new RFigure(fig_width, fig_height, POSTSCRIPT,
				      ofs+i, system, 0.5);
	    }
	    else {
	      figure[i] = new RFigure(fig_width, fig_height, BITMAP_8BIT);
	    }
	  }
	  else {
	    /* Use the same figure */
	    figure[i] = figure[0];
	  }
	}
	else if (!one_plot || i == 0) /* One plot to standard output */ {
	  /* Initialise plot */
	  if (format == FORMAT_POSTSCRIPT) {
	    figure[i] = new RFigure(fig_width, fig_height, POSTSCRIPT,
				    &cout, system, 0.5);
	  }
	  else {
	    figure[i] = new RFigure(fig_width, fig_height, BITMAP_8BIT);
	  }
	}
	else {
	  /* Use the same figure */
	  figure[i] = figure[0];
	}

	if (!one_plot || i == 0) {
	  figure[i]->clear();
	  figure[i]->axis(0,0,fig_width, fig_height);
	  if (cloud_data) {
	    figure[i]->font = font_loader->load(font_size,
						string("helvetica"),
						string("bold"),
						string("r"),
						string("adobe"));
	    if (!figure[i]->font) {
	      cerr << "Error loading helvetica bold font\n";
	    }
	    figure[i]->text(0.01, 0.0, system, ANCHOR_BOTTOMLEFT);
	  }
	  figure[i]->font = font_loader->load(font_size,
					      string("helvetica"),
					      string("medium"),
					      string("r"),
					      string("adobe"));
	  if (!figure[i]->font) {
	    cerr << "Error loading helvetica medium font\n";
	  }
	}

	y_offset[i] = y_offset_default+(height+y_offset_default)*one_plot*(max_vars-i-1);
	axes[i] = new RAxes(figure[i], x_offset, y_offset[i],
			    x_offset + width, y_offset[i] + height);

	axes[i]->axis(minx, miny, maxx, maxy);
	if (!one_plot || i == 0) {
	  if (cloud_data) {
	    get_date(tmpstr, day, month, year);
	    axes[i]->text(minx, maxy, datestr, ANCHOR_BOTTOMLEFT);
	  }
	  else {
	    axes[i]->text(0.5*(minx+maxx), maxy + ((float) (16*(maxy-miny)))/height,
			  titlestr, ANCHOR_BOTTOM);
	    axes[i]->text(0.5*(minx+maxx), maxy, paramstr, ANCHOR_BOTTOM);
	  }
	}

	// Decide what colour range to use, as well as whether it is
	// logarithmc and the tick intervals
	VarRange var_range;
	// Firstly the default settings: reflectivity ranges depending
	// if cloud or scanning data
	log_scale[i] = false;
	using_nez[i] = false;
	vfold[i] = NO_FLOAT_ATTRIBUTE;
	if (cloud_data) {
	  get_var_range("Zcloud", &var_range);
	}
	else {
	  get_var_range("Zscanning", &var_range);
	}
	minvalue[i] = var_range.min;
	maxvalue[i] = var_range.max;
	intervalue[i] = var_range.tick_interval;
	tickstartvalue[i] = var_range.tick_start;
	scale_offset = 0;
	// Next any default settings specific to this variable
	if (get_var_range(name, &var_range)) {
	  minvalue[i] = var_range.min;
	  maxvalue[i] = var_range.max;
	  intervalue[i] = var_range.tick_interval;
	  tickstartvalue[i] = var_range.tick_start;
	  log_scale[i] = var_range.logarithmic;
	}

	// Default values may be overridden by "plot_range" attribute
	plot_range = get_float_array_attribute(ncid, ivar, "plot_range",
					       &n_plot_range);
	if (plot_range && n_plot_range > 1) {
	  plot_scale = request_char_attribute(ncid, ivar, "plot_scale");
	  tickstartvalue[i] = minvalue[i] = plot_range[0];
	  if (!is_definition) {
	    maxvalue[i] = plot_range[1];
	  }
	  else {
	    maxvalue[i] = minvalue[i]+colormap.size();
	    defsize[i] = plot_range[1];
	  }
	  intervalue[i] = best_tick_spacing(minvalue[i], maxvalue[i]);
	  if (plot_scale) {
	    if (strncasecmp(plot_scale, "log", 3) == 0) {
	      log_scale[i] = true;
	      intervalue[i] = 1;
	      minvalue[i] = log10(plot_range[0]);
	      maxvalue[i] = log10(plot_range[1]);
	      /* Round to the nearest 0.5 */
	      minvalue[i] = 0.5*floor(2.0*minvalue[i]+0.5);
	      maxvalue[i] = 0.5*floor(2.0*maxvalue[i]+0.5);
	      if (minvalue[i] != floor(minvalue[i]+0.5)) {
		intervalue[i] = 0.5;
	      }
	      tickstartvalue[i] = minvalue[i];
	    }
	    delete[] plot_scale;
	  }
	  delete[] plot_range;
	}
	// Special case: lidar backscatter
	else if (strncmp(name, "beta", 4) == 0) {
	  nc_type beta_type;
	  nc_inq_vartype(ncid, ivar, &beta_type);
	  if (beta_type == NC_SHORT) {
	    /* Standard lidar file: beta is a short */
	    float scale_factor;

	    scale_offset = -7;
	    log_scale[i] = true;
	    tickstartvalue[i] = minvalue[i] = 0; maxvalue[i] = 3; intervalue[i] = 1;

	    scale_factor = get_float_attribute(ncid, ivar, "scale_factor");
	    if (scale_factor >= NO_FLOAT_ATTRIBUTE_TEST && !override_betacal) {
	      betacal = log10(scale_factor) - scale_offset;
	    }
	  }
	  else {
	    /* Cloudnet "categorization" file: beta is a float */
	    log_scale[i] = true;
	    tickstartvalue[i] = minvalue[i] = -7; maxvalue[i] = -4; intervalue[i] = 1;
	    betacal = 0.0;
	  }
	}
	// Special case: radar Doppler velocity
	else if (name[0] == 'v' && strncmp(name, "vwind", 5)) {
	  get_var_range("v", &var_range);
	  vfold[i] = get_float_attribute(ncid, ivar, "folding_velocity");
	  if (var_range.override || vfold[i] < NO_FLOAT_ATTRIBUTE_TEST) {
	    minvalue[i] = var_range.min; maxvalue[i] = var_range.max;
	    intervalue[i] = var_range.tick_interval;
	    tickstartvalue[i] = var_range.tick_start;
	  }
	  else {
	    tickstartvalue[i] = minvalue[i] = -vfold[i]; maxvalue[i] = +vfold[i];
	    intervalue[i] = best_tick_spacing(minvalue[i], maxvalue[i]);
	  }
	}
	else if (nez_calibrate && strncmp(name, "Zh", 2) == 0) {
	  if (nc_inq_varid(ncid, "nez", &nezid) == NC_NOERR) {
	    using_nez[i] = true;
	  }
	}
	else if (strlen(name) > 6 && strcmp(name+strlen(name)-5, "atten") == 0) {
	  tickstartvalue[i] = minvalue[i] = 0; maxvalue[i] = 3; intervalue[i] = 1;
	}
	/*
	else if (strncmp(name, "rainrate", 8) == 0) {
	  minvalue[i] = 0; maxvalue[i] = 50; intervalue[i] = 5;
	}
	else if (strncmp(name, "lwp", 3) == 0) {
	  minvalue[i] = -50; maxvalue[i] = 500; intervalue[i] = 50;
	}
	*/

	// Final override if a range has been set explicitly on the
	// command line
	if (get_var_range(name, &var_range) && var_range.override) {
	  minvalue[i] = var_range.min;
	  maxvalue[i] = var_range.max;
	  intervalue[i] = var_range.tick_interval;
	  tickstartvalue[i] = var_range.tick_start;
	  log_scale[i] = var_range.logarithmic;
	}
	// Or if the "-range" argument has been invoked
	else if (get_var_range("default", &var_range) && var_range.override) {
	  minvalue[i] = var_range.min;
	  maxvalue[i] = var_range.max;
	  intervalue[i] = var_range.tick_interval;
	  tickstartvalue[i] = var_range.tick_start;
	  log_scale[i] = var_range.logarithmic;
	}

	if (time_series[i]) {
	  axes[i]->axis(minx, minvalue[i], maxx, maxvalue[i]);
	}
	axes[i]->clipbox_to_axis();

	if (!no_bar) {
	  char *definition = request_char_attribute(ncid, ivar, "definition");

	  if (!time_series[i]) {
	    cbar = new RAxes(figure[i], (fig_width-10)-bar_width
			       -is_definition*definition_width,
			     y_offset[i]+0.5*(height-bar_height),
			     fig_width-10-is_definition*definition_width,
			     y_offset[i]+0.5*(height+bar_height));
	    cbar->colorbar(&colormap,VERTICAL, (int) ((definition != NULL)*(defsize[i]+1)));
	    cbar->box();
	    if (definition == NULL) {
	      cbar->axis(0,minvalue[i]+scale_offset,1,
			 maxvalue[i]+scale_offset+(definition != NULL));
	    }
	    else {
	      cbar->axis(0,minvalue[i]+scale_offset,1,
			 defsize[i]+scale_offset+(definition != NULL));
	    }
	  }
	  else {
	    cbar = axes[i];
	  }
	  if (units) {
	    if (log_scale[i]) {
	      cbar->yticks(tickstartvalue[i]+scale_offset,intervalue[i],
			   "10<SUP>%g",0);
	    }
	    else {
	      cbar->yticks(tickstartvalue[i]+scale_offset,intervalue[i],"%g",0);
	    }
	  }
       	  if (!field_title && !non_dimensional) {
	    sprintf(tmpstr,"%s (%s)", long_name, units);
	    cbar->ylabel(tmpstr);
	  }
	  else if (!field_title) {
	    cbar->ylabel(long_name);
	  }
	  else if (!non_dimensional) {
	    cbar->ylabel(units);
	  }
	  if (definition) {
	    char *ch = definition;
	    while (*ch) {
	      char *lookahead = ch;
	      int index;
	      while (*lookahead != '\n' && *lookahead != '\0') {
		lookahead++;
	      }
	      if (*lookahead == '\n') {
		*lookahead = '\0';
		lookahead++;
	      }
	      if (!sscanf(ch, "%d:\n", &index)) {
		break;
	      }
	      while (*ch != ':' && *ch != '\0') {
		ch++;
	      }
	      if (*ch) {
		ch++;
	      }
	      cbar->text(1, index+0.5, ch, ANCHOR_LEFT);
	      if (!*lookahead) {
		break;
	      }
	      ch = lookahead;
	    }
	    delete[] definition;
	  }
	  if (!time_series[i]) {
	    delete cbar;
	  }
	}

	/* Clean up */
	if (units && !units_override) {
	  delete[] units;
	  units = NULL;
	}
	if (long_name && !long_name_override) {
	  delete[] long_name;
	  long_name = NULL;
	}
      }

      if (!quiet && !standard_output) {
	cerr << "\n";
      }

      var = new float[ngates];

      } /* if firstday || ! multiday */
      else {
	/* check that all the variable IDs are right */
	for (ivar = 0; ivar < max_vars; ++ivar) {
	  handle_error(nc_inq_varid(ncid, field_name[ivar], &varid[ivar]));
	}
      }

      start[1] = 0;
      count[0] = 1;
      count[1] = ngates;

      for (iray = 0; iray < nrays; ++iray) {
	float elev, azim;

	start[0] = iray;
	handle_error(nc_get_var1_float(ncid, timeid, start, &time));

	if (time_in_s) {
	    time = time / 3600;
	}
	if (!cloud_data) {
	  handle_error(nc_get_var1_float(ncid, azimid, start, &azim));
	  handle_error(nc_get_var1_float(ncid, elevid, start, &elev));
	}
	if (iray < nrays-1) {
	  start[0] = iray+1;
	  handle_error(nc_get_var1_float(ncid, timeid, start, &nexttime));
	  if (time_in_s) {
	    nexttime = nexttime / 3600;
	  }
	  if (nexttime - time > rayduration || nexttime < time) {
	    nexttime = time + rayduration;
	  }
	  start[0] = iray;
	}
	else {
	  nexttime = time + rayduration;
	}

	if (cloud_data) {
	  /*** CLOUD DATA ***/
	  for (ivar = 0; ivar < max_vars; ++ivar) {
	    float gap = missing_value[ivar];
	    if (time_series[ivar]) {
	      float cols[2] = {0.25*(minvalue[ivar]+maxvalue[ivar]), 
			       0.25*(minvalue[ivar]+maxvalue[ivar])};
	      var[0] = 0.0;
	      handle_error(nc_get_var1_float(ncid, varid[ivar], start, var+1));
	      if (var[1] != gap && var[1] >= minvalue[ivar]) {
		if (!multiday) {
		  float dtime2 = 0.5*(nexttime-time);
		  axes[ivar]->vray(time-dtime2, nexttime-dtime2,
				   var, cols, 2, &colormap, 
				   minvalue[ivar], maxvalue[ivar], gap);
		}
		else {
		  float dtime2 = (nexttime-time)/48.0;
		  axes[ivar]->vray(day+time/24.0-dtime2, day+nexttime/24.0-dtime2,
				   var, cols, 2, &colormap, 
				   minvalue[ivar], maxvalue[ivar], gap);
		}
	      }
	      else if (var[1] == gap) {
		var[0] = minvalue[ivar];
		var[1] = maxvalue[ivar];
		cols[0] = cols[1] = minvalue[ivar];
		if (!multiday) {
		  float dtime2 = 0.5*(nexttime-time);
		  axes[ivar]->vray(time-dtime2, nexttime-dtime2,
				   var, cols, 2, &colormap, 
				   minvalue[ivar], maxvalue[ivar], gap);
		}
		else {
		  float dtime2 = (nexttime-time)/48.0;
		  axes[ivar]->vray(day+time/24.0-dtime2, day+nexttime/24.0-dtime2,
				   var, cols, 2, &colormap, 
				   minvalue[ivar], maxvalue[ivar], gap);
		}
	      }
	    }
	    else {

	    handle_error(nc_get_vara_float(ncid, varid[ivar],
					   start, count, var));
	    if (using_nez[ivar]) {
	      handle_error(nc_get_var1_float(ncid, nezid, start, &nez));
	      for (unsigned int igate = 0; igate < ngates; ++igate) {
		if (var[igate] != gap) {
		  var[igate] += (nez_override - nez);
		}
	      }	    
	    }
	    else if (log_scale[ivar]) {
	      for (unsigned int igate = 0; igate < ngates; ++igate) {
		gap = -999.0;
		if (var[igate] == missing_value[ivar]) {
		  if (plot_log_missing) {
		    var[igate] = minvalue[ivar];
		  }
		  else {
		    var[igate] = gap;
		  }
		}
		else if (var[igate] <= 0.0) {
		  var[igate] = gap;
		}
		else {
		  var[igate] = log10(var[igate]) + betacal;
		}
	      }
	    }
	    else if (vfold[ivar] > NO_FLOAT_ATTRIBUTE_TEST) {
	      for (unsigned int igate = 0; igate < ngates; ++igate) {
		float tmpvar = var[igate];
		if (tmpvar != gap && tmpvar < minvalue[ivar]
		    && tmpvar < maxvalue[ivar]) {
		  var[igate] += 2 * vfold[ivar];
		}
		else if (tmpvar != gap && tmpvar > maxvalue[ivar]
			 && tmpvar > minvalue[ivar]) {
		  var[igate] -= 2 * vfold[ivar];
		}
	      }
	    }
	    if (!multiday) {
	      float dtime2 = 0.5*(nexttime-time);
	      axes[ivar]->vray(time-dtime2, nexttime-dtime2,
			       range, var, ngates, &colormap, 
			       minvalue[ivar], maxvalue[ivar], gap);
	    }
	    else {
	      float dtime2 = (nexttime-time)/48.0;
	      axes[ivar]->vray(day+time/24.0-dtime2, day+nexttime/24.0-dtime2,
			       range, var, ngates, &colormap, 
			       minvalue[ivar], maxvalue[ivar], gap);
	    }
	    }
	  }
	}
	else {
	  /*** SCANNING DATA ***/
	  int i;
	  float x[4], y[4];
	  float tmpangle = 0;

	  switch (sweep_mode) {
	  case UF_RHI: case UF_COPLANE:
	    tmpangle = elev * PI_OVER_180;
	    if (angle < -999) {
	      dangle = 0.2 * PI_OVER_180; // default:+/-0.2 degs
	      bestangle = tmpangle;
	    }
	    else {
	      dangle = 0.7 * (tmpangle - angle);
	      bestangle = (tmpangle + angle) * 0.5;
	    }
	    angle = tmpangle;
	    break;
	  case UF_PPI:
	    if (angle < -999.0) {
	      dangle = 0.2 * PI_OVER_180; // default:+/-0.2 degs
	    }
	    else {
	    /* Note: the +90.0 is because `angle' is CLOCKWISE from DUE WEST */
	      dangle = 0.7 * ((azim+90.0) * PI_OVER_180 - angle);
	    }
	    angle = (azim+90.0) * PI_OVER_180;
	    break;
	  }

	  for (i = 0; i < ((int) ngates)-1; i++) {
	    float D0 = RADIUS_OF_THE_EARTH, D02;
	    switch (sweep_mode) {
	    case UF_RHI: case UF_COPLANE:
	      /* Correct for the curvature of the earth */
	      D02 = D0 * D0;
	      x[0] = range[i]*cos(bestangle + dangle);
	      y[0] = range[i]*sin(bestangle + dangle)
		+ sqrt(D02 + x[0]*x[0]) - D0;
	      x[1] = range[i+1]*cos(bestangle + dangle);
	      y[1] = range[i+1]*sin(bestangle + dangle)
		+ sqrt(D02 + x[1]*x[1]) - D0;
	      x[2] = range[i+1]*cos(bestangle - dangle);
	      y[2] = range[i+1]*sin(bestangle - dangle)
		+ sqrt(D02 + x[2]*x[2]) - D0;
	      x[3] = range[i]*cos(bestangle - dangle);
	      y[3] = range[i]*sin(bestangle - dangle)
		+ sqrt(D02 + x[3]*x[3]) - D0;
	      break;
	    case UF_PPI:
	      x[0] = -range[i]*cos(angle + dangle);
	      y[0] = range[i]*sin(angle + dangle);
	      x[1] = -range[i+1]*cos(angle + dangle);
	      y[1] = range[i+1]*sin(angle + dangle);
	      x[2] = -range[i+1]*cos(angle - dangle);
	      y[2] = range[i+1]*sin(angle - dangle);
	      x[3] = -range[i]*cos(angle - dangle);
	      y[3] = range[i]*sin(angle - dangle);
	      break;
	    default:
	      x[0] = x[1] = (time-start_time) * 60.0;
	      x[2] = x[3] = (nexttime-start_time) * 60.0;
	      y[0] = y[3] = range[i];
	      y[1] = y[2] = range[i+1];
	    }
	    for (ivar = 0; ivar < max_vars; ivar++) {
	      float value;
	      size_t index[2] = {iray, i};
	      int j;

	      handle_error(nc_get_var1_float(ncid, varid[ivar], index, &value));
	      if (value != missing_value[ivar]) {
		if (log_scale[ivar]) {
		  value = log10(value);
		}
		j = (int) (((float) colormap.size())
			   * (value - minvalue[ivar])
			   / (maxvalue[ivar] - minvalue[ivar]));
		if (j < 0) j = 0;
		if (j > colormap.size()-1) j = colormap.size()-1;
		axes[ivar]->polygon(4, x, y, colormap.get(j));
	      }
	    }
	    /* Optimisation for RHIs: skip rest of ray if y[0] out of range */
	    if (sweep_mode != UF_PPI && (x[0] > maxx+1.0 || y[0] > maxy+1.0)) {
	      break;
	    }
	  }
	}
      }

      if (!multiday || lastday) {

      delete[] var;
      delete[] range; 
      if (system_true) {
	if (*system_true)
	  delete[] system_true;
	system_true = NULL;
      }

      dhour = (endhour-starthour > 6) ? floor((endhour-starthour)
					      / 6.0) :
		(1.0+floor((endhour-starthour)/4.0))/2.0;
      for (ivar = 0; ivar < max_vars; ++ivar) {
       	axes[ivar]->box();
	if (cloud_data) {
	  if (!multiday) {
	    axes[ivar]->xticks_time(0, dhour);
	  }
	  else {
	    axes[ivar]->xticks(0, 1);
	  }
	  if (!multiday) {
	    axes[ivar]->xlabel("Time (UTC)");
	  }
	  else {
	    axes[ivar]->xlabel("Day of the month");
	  }
	  if (!time_series[ivar]) {
	    axes[ivar]->yticks(0,1);
	    axes[ivar]->ylabel("Height (km)");
	  }
	  else {
	    axes[ivar]->yticks(tickstartvalue[ivar], intervalue[ivar], "%g", 0);
	  }
	}
	else {
	  switch (sweep_mode) {
	  case UF_RHI: case UF_COPLANE:
	    axes[ivar]->xticks(floor(minx*0.1)*10,10);
	    axes[ivar]->yticks(floor(miny*0.5)*2,2);
	    axes[ivar]->xlabel("Distance (km)");
	    axes[ivar]->ylabel("Height (km)");
	    break;
	  case UF_PPI:
	    axes[ivar]->xticks(floor(minx*0.1)*10,10);
	    axes[ivar]->yticks(floor(miny*0.1)*10,10);
	    axes[ivar]->xlabel("Distance east (km)");
	    axes[ivar]->ylabel("Distance north (km)");
	    if (mapfile) {
	      figure[ivar]->clipbox(axes[ivar]->position());
	      MAPPlotCoastline(axes[ivar]);
	      figure[ivar]->clipbox(false);
	    }
	    if (locfile) {
	      figure[ivar]->clipbox(axes[ivar]->position());
	      MAPPlotLocations(axes[ivar]);
	      figure[ivar]->clipbox(false);
	    }
	    break;
	  default:
	    axes[ivar]->xticks(floor(minx), 1);
	    axes[ivar]->yticks(floor(miny*0.5)*2,2);
	    axes[ivar]->xlabel("Approximate time (minutes)");
	    axes[ivar]->ylabel("Range (km)");
	  }
	}
      }
      if (field_title) {
	for (ivar = 0; ivar < max_vars; ++ivar) {
	  if (long_name_override) {
	    long_name = long_name_override;
	  }
	  else {
	    long_name = get_char_attribute(ncid, varid[ivar], "long_name");
	  }
	  figure[ivar]->font = font_loader->load(font_size,
						 string("helvetica"),
						 string("bold"),
						 string("r"),
						 string("adobe"));
	  figure[ivar]->text(x_offset+4, y_offset[ivar]+height-2,
			     long_name, ANCHOR_TOPLEFT);
	  if (long_name && !long_name_override) {
	    delete[] long_name;
	    long_name = NULL;
	  }
	}
      }

      for (ivar = 0; ivar < max_vars; ++ivar) {
	if (format == FORMAT_PNG && (!one_plot || ivar == 0)) {
	  if (!standard_output) {
	    figure[ivar]->write_png(&ofs[ivar]);
	  }
	  else {
	    figure[ivar]->write_png(&cout);
	  }
	}
	delete axes[ivar];
	if (!one_plot || ivar == 0) {
	  delete figure[ivar];
	}
	if (!standard_output) {
	  ofs[ivar].close();
	}
      }
      if (standard_output) {
	break;
      }
      }

      handle_error(nc_close(ncid));
    }
    delete font_loader;
  }
  catch (const char *error_message) {
    cerr << error_message << endl;
    exit(1);
  }
}
