/* 
   Copyright (C) 2000 Robin Hogan

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <fstream>
#include <iomanip>
#include <string>
#include <cmath>
#include <cstdio>
#include "rplot.h"

// *** Font class ***

// *** RRectangle class ***

ostream&
operator<<(ostream& os, const RRectangle &rect)
{
  os << "(" << rect.x1 << "," << rect.y1 
     << " ; " << rect.x2 << "," << rect.y2 << ")";
  return os;
}


// *** RColor class ***

ostream&
operator<<(ostream& os, const RColor &col)
{
  os << "(" << (int) col.red << "," << (int) col.green 
     << "," << (int) col.blue << "," << (int) col.alpha << ")";
  return os;
}

// *** RPalette class ***

void
RPalette::clear(void)
{
  int i;
  RColor null_color=make_color(0,0,0,0);

  _size=0;_next=0;_last=0;_max=0;
  for (i=0;i<MAX_COLORS;i++) {
    _status[i]=0;
    _color[i]=null_color;
  } 
}

RColor
RPalette::set(const uchar index, const RColor &col)
{
  if (_status[index]==false) {
    _status[index]=true;
    _size++;
    if (_max<index) {
      _max=index;
    }
    if (index < MAX_COLORS && _status[index+1] == false) {
      _next=index+1;
    }
    else if (_size < MAX_COLORS) {
      _next=0;
      while (_status[_next]==true && _next<MAX_COLORS) _next++;
    }
  }
  _last=index;
  return _color[index]=col;
}

void
RPalette::setrange(const int i1, const int i2, const RColor &col1,
		   const RColor &col2)
{
  int i, isize;

  if ((isize=i2-i1) == 0) {
    if (i1>=0 && i1<MAX_COLORS) {
      set(i1,make_color( (col2.red + col1.red)/2, (col2.green + col1.green)/2, 
			 (col2.blue + col1.blue/2) ));
    }
  }
  else if (isize > 0) {
    for (i=i1;i<=i2;i++) {
      if (i>=0 && i<MAX_COLORS) {
	set(i,make_color( (col2.red * (i-i1) + col1.red * (i2-i))/isize,
			  (col2.green * (i-i1) + col1.green * (i2-i))/isize,
			  (col2.blue * (i-i1) + col1.blue * (i2-i))/isize));
      }
    }
  }
}

/*
RColor& RPalette::operator[](const uchar index)  {
	if (_status[index]==0) {
		_color[index]=RColor::RColor(0,0,0);
		_status[index]=1;
		_size++;
		if (_max<index) {
			_max=index;
		}
		if (index < MAX_COLORS && _status[index+1] == 0) {
			_next=index+1;
		}
		else if (_size < MAX_COLORS) {
			_next=0;
			while (_status[_next]==1 && _next<MAX_COLORS) _next++;
		}
	}
	_last=index;
	return _color[index];
}
*/

uchar
RPalette::operator+=(const RColor &col)
{
  int i;
  if (equal(_last,col)) {
    return _last;
  }
  else if (_size<MAX_COLORS) {
    for (i=(_max==_last ? (_max-1) : _max) ; i>=0 ; i--) {
      if (equal((uchar) i,col)) {
	return _last=i;
      }
    }
    set(_next,col);
    return _last;
  }
  else {
    int diff,mindiff=4*MAX_COLORS*MAX_COLORS,idiff=0;
    for (i=0 ; i<MAX_COLORS ; i++) {
      diff = _color[i]-col;
      if (mindiff>diff) {
	mindiff=diff;
	idiff=i;
      }
    }
    return _last=i;
  }
}

int
RPalette::load(const char *filename)
{
  ifstream ifs(filename);
  int red, green, blue;
  int n = 0;

  if (!ifs) {
    return 1;
  }

  while (n < MAX_COLORS && ifs >> red >> green >> blue) {
    if (red < 0) red = 0; else if (red > 255) red = 255;
    if (green < 0) green = 0; else if (green > 255) green = 255;
    if (blue < 0) blue = 0; else if (blue > 255) blue = 255;
    set(n, make_color(red, green, blue));
    ++n;
  }

  ifs.close();
  return 0;
}

int
RPalette::write(const char *filename)
{
  ofstream ofs(filename);
  int n = 0;

  if (!ofs) {
    return 1;
  }

  for (n = 0; n <= _max; ++n) {
    if (_status[n]) {
      ofs << setw(4) << (int) _color[n].red << " "
	  << setw(4) << (int) _color[n].green << " "
	  << setw(4) << (int) _color[n].blue << "\n";
    }
  }
  ofs.close();
  return 0;
}

ostream&
operator<<(ostream& os, const RPalette &pal)
{
  int i;
  os << "Size=" << (int) pal.size() << "  Max=" << (int) pal.max() << 
    "  Contiguous=" << (int) pal.contiguous() << "\n";
  for (i=0;i<MAX_COLORS;i++) {
    if (pal.status(i)) {
      os << i << ": " << pal.get(i) << "\n";
    }
    else {
      os << i << ": not assigned\n";
    }
  }
  return os;
}

void
best_axis_range(double *a1, double *a2, double *da, double *na)
{
  double b1=0, b2=0, db=1, nb=5;
  double ra=(*a2)-(*a1), scale;
  if (ra<0) {
    best_axis_range(a2,a1,da,na);
    return;
  }
  else if (ra==0) {
    (*a1) += 1;
    (*a2) -= 1;
    (*da) = 1.0;
    (*na) = 2.0;
    return;
  }
  if ((*da) == 0) {
    if ((*na) != 0) {
      nb=ABS(*na);
    }
    db=ra/nb;
    scale=pow(10,floor(log10(db)));
    if (db/scale < 2) {
      db=2*scale;
    }
    else if (db/scale < 5) {
      db=5*scale;
    }
    else {
      db=10*scale;
    }
  }
  else {
    db=(*da);
  }
  b1=db*(floor((*a1)/db));
  b2=db*(ceil((*a2)/db));
  nb=ceil((b2-b1)/db);
  if ((*na) > nb) {
    b1=b1-db*floor(0.501*((*na)-nb));
    nb=(*na);
    b2=b1+nb*db;
  }
  *a1=b1;
  *a2=b2;
  *da=db;
  *na=nb;
  return;
}

float
best_tick_spacing(float x1, float x2, int n_ticks)
{
  float log_ideal_spacing = log10((x2-x1)/n_ticks);
  float order_of_magnitude = floor(log_ideal_spacing);
  float log_remnant = log_ideal_spacing-order_of_magnitude;
  int ibase = 1;
  float base_spacing[] = {0.5, 1, 2, 2.5, 5, 10, 20};
  if (log_remnant < 0.15) {
    ibase = 1;
  }
  else if (log_remnant < 0.35) {
    ibase = 2;
  }
  else if (log_remnant < 0.55) {
    ibase = 3;
  }
  else if (log_remnant < 0.85) {
    ibase = 4;
  }
  else {
    ibase = 5;
  }

  if (x1 != 0.0) {
    float x1_remnant = fabs(x1)/pow(10.0,order_of_magnitude);
    int round_x1_remnant = (int) floor(x1_remnant+0.5);
    if (fabs(x1_remnant - round_x1_remnant) < 0.05) {
      /* We have a round number */
      if (fmod(round_x1_remnant, base_spacing[ibase]) > 0.05) {
	/* Current base spacing wouldn't permit a zero value */
	ibase--;
      }
    }
  }
  return pow(10.0, order_of_magnitude) * base_spacing[ibase];
}

void
create_jet_palette(RPalette *pal,uchar num)
{
  RColor
    lightgrey(223,223,223),
    lightblue(127,127,255),
    blue(0,0,255),
    turquoise(0,159,159), 
    green(0,255,0),
    yellow(255,255,0),
    red(255,0,0),
    darkmagenta(127,0,127);
  int n=num;
  pal->setrange(0,n/8-1,lightgrey,lightblue);
  pal->setrange(n/8-1,n/4-1,lightblue,blue);
  pal->setrange(n/4-1,(n*3)/8-1,blue,turquoise);
  pal->setrange((n*3)/8-1,n/2-1,turquoise,green);
  pal->setrange(n/2-1,(n*5)/8-1,green,yellow);
  pal->setrange((n*5)/8-1,(n*7)/8-1,yellow,red);
  pal->setrange((n*7)/8-1,n-1,red,darkmagenta);
}

void
create_litmus_palette(RPalette *pal,uchar num)
{
  RColor
    darkgreen(0,63,0),
    darkblue(0,0,128),
    blue(0,0,255),
    cyan(0,255,255), 
    lightgrey(223,223,223),
    yellow(255,255,0),
    red(0,255,0,0),
    darkmagenta(127,0,127);
  int n=num;
  pal->setrange(0,n/8-1,darkgreen,darkblue);
  pal->setrange(n/8-1,n/4-1,darkblue,blue);
  pal->setrange(n/4-1,n/2-3,blue,cyan);
  pal->setrange(n/2-2,n/2-1,lightgrey,lightgrey);
  pal->setrange(n/2,(n*3)/4-1,yellow,red);
  pal->setrange((n*3)/4-1,n-1,red,darkmagenta);
}

/*
void
create_cloud_palette(RPalette *pal,uchar num)
{
  RColor 
    darkblue(0,0,128),
    turquoise(0, 159, 159),
    blue(0,0,255),
    white(255,255,255),
    grey(63,63,63);
  int n=num;
  pal->setrange(0,n-1,turquoise,white);
  //  pal->setrange(n/2,(n*3)/4-1,blue,white);
  //  pal->setrange((n*3)/4,n-1,white,grey);
  //  pal->setrange(n/2,n-1,white,grey);
}
*/
