#include <math.h>
#include <stdio.h>
#include "zdr_phidp.h"

/* Forward model to predict the differential reflectivity Zdr and
   differential phase shift, together with their Jacobians, from input
   rays of reflectivity Zh and coefficients a and b (in the
   relationship Zh=a*R^b) */
int
forward_model(ScatteringLUT *lut, int n, OEreal drange, OEreal *range,
	      OEreal *temperature,
	      OEreal *Zh_lin, OEreal *ln_a, OEreal b,
	      OEreal *hail_fraction,
	      OEreal atten_scaling, OEreal max_atten,
	      /* Output */
	      OEreal *Zdr, OEreal *phi_dp,
	      OEreal *Ah, OEreal *Av,
	      OEreal **dZdr_dlna, OEreal **dphi_dp_dlna,
	      OEreal **dAh_dlna, OEreal **dAv_dlna,
	      OEreal **dZdr_dhf, OEreal **dphi_dp_dhf,
	      OEreal **dAh_dhf, OEreal **dAv_dhf,
	      OEreal *dZdr_das, OEreal *dphi_dp_das,
	      OEreal *dAh_das, OEreal *dAv_das)
{
  int i;
  int is_hail = 0;
  int reached_max_atten = 0;
  OEreal Afactor = drange * 0.2 * log(10.0);
  OEreal Bfactor = 0.1 * log(10.0);
  if (hail_fraction) {
    is_hail = 1;
  }

  Ah[0] = 0.0;
  Av[0] = 0.0;
  if (dAh_das && dAv_das) {
    dAh_das[0] = dAv_das[0] = 0.0;
  }
  if (phi_dp) {
    phi_dp[0] = 0.0;
  }
  if (dZdr_das) {
    dZdr_das[0] = 0.0;
  }
  if (dphi_dp_das) {
    dphi_dp_das[0] = 0.0;
  }

  //  OE_CHECK(oe_print_matrix(stderr, &lut->Zh_Kdp, "%(LUT) Zh_Kdp"));
  /* Loop through each range gate */
  for (i = 0; i < n; i++) {
    OEreal Zh_unat, Zh_rain, Zh_hail, hf = 0.0, ln_Zh_R;
    OEreal weight, weight1, weight2;
    OEreal Zh_Ah_lut, Zh_Av_lut, Zh_Kdp_lut;
    OEreal Zdr_intrinsic = 0.0, Zdr_rain_intrinsic = 0.0;
    OEreal dZh_Ah_dln_Zh_R, dZh_Av_dln_Zh_R, dZdr_dln_Zh_R, dZh_Kdp_dln_Zh_R;
    OEreal dAh_dZh, dAv_dZh, dKdp_dZh;
    int iint = 0, iT = 0;
    int j;

    if (is_hail) {
      hf = hail_fraction[i];
    }
    /* Calculate the linear unattenuated reflectivity using the total
       2-way attenuation up to that point, Ah */
    Zh_unat = Zh_lin[i] * pow(10.0, 0.1*Ah[i]);
    Zh_rain = Zh_unat * (1-hf);
    Zh_hail = Zh_unat * hf;
    /* Calculate the log of the ratio of Zh to rainrate */
    ln_Zh_R = (1-1/b)*log(Zh_rain) + ln_a[i]/b;

    /* Nearest-neighbour interpolation for temperature */
    while (iT < lut->temperature.n-1
	   && temperature[i] > 0.5*(lut->temperature.data[iT]
				    +lut->temperature.data[iT+1])) {
      iT++;
    }

    /* Find the point in the look-up table */
    while (ln_Zh_R > lut->ln_Zh_R.data[iT][iint]
	   &&  iint < lut->ln_Zh_R.n-2) {
      iint++;
    }

    /* Calculate the weights */
    weight  = 1.0 / (lut->ln_Zh_R.data[iT][iint+1]
		     - lut->ln_Zh_R.data[iT][iint]);
    weight1 = (lut->ln_Zh_R.data[iT][iint+1] - ln_Zh_R) * weight;
    weight2 = (ln_Zh_R - lut->ln_Zh_R.data[iT][iint]) * weight;

    /* Perform the interpolations */
    if (Zdr) {
      Zdr_rain_intrinsic = lut->Zdr.data[iT][iint] * weight1
	+ lut->Zdr.data[iT][iint+1] * weight2;
      /* Calculate the Zdr including differential attenuation */
      if (hf > 0.0) {
	/* CHANGE 22/12/2005: hf/lut->hail_Zdr_lin */
	/*	Zdr_intrinsic = -10.0*log10(hf*lut->hail_Zdr_lin
		+ (1.0-hf)*pow(10.0, -0.1*Zdr_rain_intrinsic));*/
	Zdr_intrinsic = -10.0*log10(hf/lut->hail_Zdr_lin
		    + (1.0-hf)*pow(10.0, -0.1*Zdr_rain_intrinsic));
      }
      else {
	Zdr_intrinsic = Zdr_rain_intrinsic;
      }
      Zdr[i] = Zdr_intrinsic - Ah[i] + Av[i];
    }
    if (i >= n-1) {
      /* Final loop only for Zdr: quit before further calculations */
      break;
    }

    /* Continue the interpolations */
    Zh_Ah_lut = exp(lut->ln_Zh_attenh.data[iT][iint] * weight1
      + lut->ln_Zh_attenh.data[iT][iint+1] * weight2)/atten_scaling;
    Zh_Av_lut = exp(lut->ln_Zh_attenv.data[iT][iint] * weight1
      + lut->ln_Zh_attenv.data[iT][iint+1] * weight2)/atten_scaling;
    
    /* Calculate the corresponding gradients */
    dZh_Ah_dln_Zh_R = Zh_Ah_lut * weight
      * (lut->ln_Zh_attenh.data[iT][iint+1]-lut->ln_Zh_attenh.data[iT][iint]);
    dZh_Av_dln_Zh_R = Zh_Av_lut * weight
      * (lut->ln_Zh_attenv.data[iT][iint+1]-lut->ln_Zh_attenv.data[iT][iint]);
    dZdr_dln_Zh_R = weight
      * (lut->Zdr.data[iT][iint+1] - lut->Zdr.data[iT][iint]);

    Zh_Kdp_lut = exp(lut->ln_Zh_Kdp.data[iT][iint] * weight1
		     + lut->ln_Zh_Kdp.data[iT][iint+1] * weight2);

    dZh_Kdp_dln_Zh_R = Zh_Kdp_lut * weight
      * (lut->ln_Zh_Kdp.data[iT][iint+1] - lut->ln_Zh_Kdp.data[iT][iint]);

    /* Calculate the total 2-way attenuations and phase shift in the
       next gate */
    if (Zh_Ah_lut > 0.0) {
      OEreal new_atten = 2.0 * drange * Zh_rain / Zh_Ah_lut;
      if (Ah[i] + new_atten > max_atten) {
	reached_max_atten = 1;
	Ah[i+1] = Ah[i];
      }
      else {
	Ah[i+1] = Ah[i] + new_atten;
      }
    }
    else {
      Ah[i+1] = Ah[i];
    }
    if (Zh_Av_lut > 0.0 && !reached_max_atten) {
      Av[i+1] = Av[i] + 2.0 * drange * Zh_rain / Zh_Av_lut;
    }
    else {
      Av[i+1] = Av[i];
    }
    /* Include the attenuation of the hail */
    if (hf > 0.0 && !reached_max_atten) {
      Ah[i+1] += 2.0 * drange * Zh_hail * lut->hail_Ah_Zh;
      Av[i+1] += 2.0 * drange * Zh_hail * lut->hail_Av_Zh;
    }

    if (phi_dp) {
      if (Zh_Kdp_lut > 0.0) {
	phi_dp[i+1] = phi_dp[i] + 2.0 * drange * Zh_rain / Zh_Kdp_lut;
      }
      else {
	phi_dp[i+1] = phi_dp[i];
      }
      /* Include the phase shift of the hail */
      if (hf > 0.0) {
	phi_dp[i+1] += 2.0 * drange * Zh_hail * lut->hail_Kdp_Zh;
      }
    }


    /*
    fprintf(stderr, "! %d %g %g %g %g %g %g\n", iint, ln_Zh_R,
	    lut->Zh_Kdp.data[iT][iint], weight1,
 	    lut->Zh_Kdp.data[iT][iint+1], weight2, weight); 
    fprintf(stderr, "* %g %g %g %g %g\n",
	    phi_dp[i+1], phi_dp[i], drange, Zh_unat, Zh_Kdp_lut);
    */

    /* Intermediate derivatives */
    if (!reached_max_atten) {
      /* After maximum attenuation reached, the derivatives need to
	 reflect the fact that Zh at this gate is no longer being
	 reduced by attenuation at an earlier gate */
      dAh_dZh = 1.0/Zh_Ah_lut
	- dZh_Ah_dln_Zh_R*(1-1/b)/(Zh_Ah_lut*Zh_Ah_lut);
      dAv_dZh = 1.0/Zh_Av_lut
	- dZh_Av_dln_Zh_R*(1-1/b)/(Zh_Av_lut*Zh_Av_lut);
      Zh_hail = 0.0;
    }
    else {
      dAh_dZh = 0.0;
      dAv_dZh = 0.0;
    }
    dKdp_dZh = 1.0/Zh_Kdp_lut
      - dZh_Kdp_dln_Zh_R*(1-1/b)/(Zh_Kdp_lut*Zh_Kdp_lut);

    /* Derivatives with respect to ln(a) */
    /* Diagonal and first off-diagonal elements of the Jacobian */
    if (dphi_dp_dlna) {
      dphi_dp_dlna[i+1][i]
	= -2.0*drange*dZh_Kdp_dln_Zh_R*Zh_rain/(b*Zh_Kdp_lut*Zh_Kdp_lut);
    }
    dAh_dlna[i+1][i]
      = -2.0*drange*dZh_Ah_dln_Zh_R*Zh_rain/(b*Zh_Ah_lut*Zh_Ah_lut);
    dAv_dlna[i+1][i]
      = -2.0*drange*dZh_Av_dln_Zh_R*Zh_rain/(b*Zh_Av_lut*Zh_Av_lut);
    if (dZdr_dlna) {
      if (hf > 0.0) {
	dZdr_dlna[i][i] = (dZdr_dln_Zh_R/b) * (1.0-hf)
	  * pow(10.0, 0.1*(Zdr_intrinsic-Zdr_rain_intrinsic));
      }
      else {
	dZdr_dlna[i][i] = dZdr_dln_Zh_R/b;
      }
    }
    /* Upper triangular elements */
    if (dphi_dp_dlna) {
      for (j = 0; j < i; j++) {
	dphi_dp_dlna[i+1][j] = dphi_dp_dlna[i][j]
	  + dAh_dlna[i][j] * Afactor * (dKdp_dZh*Zh_rain
					+ lut->hail_Kdp_Zh*Zh_hail);
      }
    }
    if (dZdr_dlna) {
      for (j = 0; j < i; j++) {      
	dAh_dlna[i+1][j] = dAh_dlna[i][j] * (1.0 + Afactor
		     * (dAh_dZh*Zh_rain + lut->hail_Ah_Zh*Zh_hail));
	dAv_dlna[i+1][j] = dAv_dlna[i][j] * (1.0 + Afactor
		     * (dAv_dZh*Zh_rain + lut->hail_Av_Zh*Zh_hail));
	dZdr_dlna[i+1][j] = - dAh_dlna[i+1][j] + dAv_dlna[i+1][j]
	  + Bfactor*(1.0-hf)*dZdr_dln_Zh_R*dAh_dlna[i+1][j]*(1.0-1.0/b)
	  *pow(10.0, 0.1*(Zdr_intrinsic-Zdr_rain_intrinsic));
      }
    }

    /* Derivatives with respect to the attenuation scaling factor */
    /* First off-diagonal elements of the Jacobian */
    if (dAh_das && dAv_das) {
      if (!reached_max_atten) {
	dAh_das[i+1] = dAh_das[i]
	  + 2.0*drange*(Zh_rain/Zh_Ah_lut + Zh_hail*lut->hail_Ah_Zh)
	  * (1.0 + 0.2303*atten_scaling*dAh_das[i]);
	dAv_das[i+1] = dAv_das[i]
	  + 2.0*drange*(Zh_rain/Zh_Av_lut + Zh_hail*lut->hail_Av_Zh)
	  * (1.0 + 0.2303*atten_scaling*dAh_das[i]);
      }
      else {
	dAh_das[i+1] = dAh_das[i];
	dAv_das[i+1] = dAv_das[i];
      }
      if (dZdr_das) {
	dZdr_das[i+1] = - dAh_das[i+1] + dAv_das[i+1];
      }
      if (dphi_dp_das) {
	if (!reached_max_atten) {
	  dphi_dp_das[i+1] = dphi_dp_das[i]
	    + 2.0*drange*(Zh_rain/Zh_Kdp_lut + Zh_hail*lut->hail_Kdp_Zh)
	    * 0.2303*atten_scaling*dAh_das[i];
	}
	else {
	  dphi_dp_das[i+1] = dphi_dp_das[i];
	}
      }
    }

    /* Derivatives with respect to the hail fraction*/
    if (is_hail) {
      /* First off-diagonal elements of the Jacobian */
      if (dphi_dp_dhf) {
	dphi_dp_dhf[i+1][i]
	  = 2.0*drange*Zh_unat*(lut->hail_Kdp_Zh - 1.0/Zh_Kdp_lut);
      }
      dAh_dhf[i+1][i]
	= 2.0*drange*Zh_unat*(lut->hail_Ah_Zh - 1.0/Zh_Ah_lut);
      dAv_dhf[i+1][i]
	= 2.0*drange*Zh_unat*(lut->hail_Av_Zh - 1.0/Zh_Av_lut);
    }
    if (dZdr_dhf) {
      dZdr_dhf[i][i] = (pow(10.0, -0.1*Zdr_rain_intrinsic)
			-lut->hail_Zdr_lin)
	* 4.343 * pow(10.0, 0.1*Zdr_intrinsic);
    }
    /* Upper triangular elements */
    if (dphi_dp_dhf) {
      for (j = 0; j < i; j++) {
	dphi_dp_dhf[i+1][j] = dphi_dp_dhf[i][j]
	  + dAh_dhf[i][j] * Afactor * (dKdp_dZh*Zh_rain
				       + lut->hail_Kdp_Zh*Zh_hail);
      }
    }
    if (dZdr_dhf) {
      for (j = 0; j < i; j++) {      
	dAh_dhf[i+1][j] = dAh_dhf[i][j] * (1.0 + Afactor
		     * (dAh_dZh*Zh_rain + lut->hail_Ah_Zh*Zh_hail));
	dAv_dhf[i+1][j] = dAv_dhf[i][j] * (1.0 + Afactor
		     * (dAv_dZh*Zh_rain + lut->hail_Av_Zh*Zh_hail));
	dZdr_dhf[i+1][j] = - dAh_dhf[i+1][j] + dAv_dhf[i+1][j];
      } 
    }
  }

  return OE_SUCCESS;
}
  
  
