#include <math.h>
#include <stdio.h>
#include "zdr_phidp.h"

/* Forward model to predict the differential reflectivity Zdr and
   differential phase shift, together with their Jacobians, from input
   rays of reflectivity Zh and coefficients a and b (in the
   relationship Zh=a*R^b) */
int
forward_model(ScatteringLUT *lut, int n, OEreal drange, OEreal *range,
	      OEreal *temperature,
	      OEreal *Zh_lin, OEreal *ln_a, OEreal b,
	      /* Output */
	      OEreal *Zdr, OEreal *phi_dp,
	      OEreal *Ah, OEreal *Av,
	      OEreal **dZdr_dlna, OEreal **dphi_dp_dlna,
	      OEreal **dAh_dlna, OEreal **dAv_dlna)
{
  int i;

  Ah[0] = 0.0;
  Av[0] = 0.0;
  if (phi_dp) {
    phi_dp[0] = 0.0;
  }

  //  OE_CHECK(oe_print_matrix(stderr, &lut->Zh_Kdp, "%(LUT) Zh_Kdp"));
  /* Loop through each range gate */
  for (i = 0; i < n; i++) {
    /* Calculate the linear unattenuated reflectivity using the total
       2-way attenuation up to that point, Ah */
    OEreal Zh_unat = Zh_lin[i] * pow(10.0, 0.1*Ah[i]);
    /* Calculate the log of the ratio of Zh to rainrate */
    OEreal ln_Zh_R = (1-1/b)*log(Zh_unat) + ln_a[i]/b;

    OEreal weight, weight1, weight2, factor;
    OEreal Zh_Ah_lut, Zh_Av_lut, Zdr_lut, Zh_Kdp_lut;
    OEreal dZh_Ah_dln_Zh_R, dZh_Av_dln_Zh_R, dZdr_dln_Zh_R, dZh_Kdp_dln_Zh_R;
    OEreal dAh_dZh, dAv_dZh, dKdp_dZh;
    int iint = 0, iT = 0;
    int j;

    /* Nearest-neighbour interpolation for temperature */
    while (temperature[i] < lut->temperature.data[iT]
	   &&  iT < lut->temperature.n-2) {
      iT++;
    }
    if (fabs(temperature[i] - lut->temperature.data[iT])
	> fabs(lut->temperature.data[iT+1] - temperature[i])) {
      iT++;
    }

    /* Find the point in the look-up table */
    while (ln_Zh_R > lut->ln_Zh_R.data[iT][iint]
	   &&  iint < lut->ln_Zh_R.n-2) {
      iint++;
    }

    /* Calculate the weights */
    weight  = 1.0 / (lut->ln_Zh_R.data[iT][iint+1]
		     - lut->ln_Zh_R.data[iT][iint]);
    weight1 = (lut->ln_Zh_R.data[iT][iint+1] - ln_Zh_R) * weight;
    weight2 = (ln_Zh_R - lut->ln_Zh_R.data[iT][iint]) * weight;

    /* Perform the interpolations */
    if (Zdr) {
      Zdr_lut = lut->Zdr.data[iT][iint] * weight1
	+ lut->Zdr.data[iT][iint+1] * weight2;
      /* Calculate the Zdr including differential attenuation */
      Zdr[i] = Zdr_lut - Ah[i] + Av[i];
    }
    if (i >= n-1) {
      /* Final loop only for Zdr: quit before further calculations */
      break;
    }
    //    OE_DEBUG;
    /* Continue the interpolations */
    Zh_Ah_lut = exp(lut->ln_Zh_attenh.data[iT][iint] * weight1
      + lut->ln_Zh_attenh.data[iT][iint+1] * weight2);
    Zh_Av_lut = exp(lut->ln_Zh_attenv.data[iT][iint] * weight1
      + lut->ln_Zh_attenv.data[iT][iint+1] * weight2);
    
    /* Calculate the corresponding gradients */
    dZh_Ah_dln_Zh_R = Zh_Ah_lut * weight
      * (lut->ln_Zh_attenh.data[iT][iint+1]-lut->ln_Zh_attenh.data[iT][iint]);
    dZh_Av_dln_Zh_R = Zh_Av_lut * weight
      * (lut->ln_Zh_attenv.data[iT][iint+1]-lut->ln_Zh_attenv.data[iT][iint]);
    dZdr_dln_Zh_R = weight
      * (lut->Zdr.data[iT][iint+1] - lut->Zdr.data[iT][iint]);

    Zh_Kdp_lut = exp(lut->ln_Zh_Kdp.data[iT][iint] * weight1
		     + lut->ln_Zh_Kdp.data[iT][iint+1] * weight2);

    dZh_Kdp_dln_Zh_R = Zh_Kdp_lut * weight
      * (lut->ln_Zh_Kdp.data[iT][iint+1] - lut->ln_Zh_Kdp.data[iT][iint]);

    /* Calculate the total 2-way attenuations and phase shift in the
       next gate */
    if (Zh_Ah_lut > 0.0) {
      Ah[i+1] = Ah[i] + 2.0 * drange * Zh_unat / Zh_Ah_lut;
    }
    else {
      Ah[i+1] = Ah[i];
    }
    if (Zh_Av_lut > 0.0) {
      Av[i+1] = Av[i] + 2.0 * drange * Zh_unat / Zh_Av_lut;
    }
    else {
      Av[i+1] = Av[i];
    }
    if (phi_dp) {
      if (Zh_Kdp_lut > 0.0) {
	phi_dp[i+1] = phi_dp[i] + 2.0 * drange * Zh_unat / Zh_Kdp_lut;
      }
      else {
	phi_dp[i+1] = phi_dp[i];
      }
    }
    /*
    fprintf(stderr, "! %d %g %g %g %g %g %g\n", iint, ln_Zh_R,
	    lut->Zh_Kdp.data[iT][iint], weight1,
 	    lut->Zh_Kdp.data[iT][iint+1], weight2, weight); 
    fprintf(stderr, "* %g %g %g %g %g\n",
	    phi_dp[i+1], phi_dp[i], drange, Zh_unat, Zh_Kdp_lut);
    */
    factor = drange * 0.2 * log(10.0) * Zh_unat;

    /* Intermediate derivatives */
    dAh_dZh = 1.0/Zh_Ah_lut
      - dZh_Ah_dln_Zh_R*(1-1/b)/(Zh_Ah_lut*Zh_Ah_lut);
    dAv_dZh = 1.0/Zh_Av_lut
      - dZh_Av_dln_Zh_R*(1-1/b)/(Zh_Av_lut*Zh_Av_lut);
    dKdp_dZh = 1.0/Zh_Kdp_lut
      - dZh_Kdp_dln_Zh_R*(1-1/b)/(Zh_Kdp_lut*Zh_Kdp_lut);

    /* Diagonal and first off-diagonal elements of the Jacobian */
    if (dphi_dp_dlna) {
      dphi_dp_dlna[i][i] = 0.5 * ((dphi_dp_dlna[i+1][i] 
	   = -2.0*drange*dZh_Kdp_dln_Zh_R*Zh_unat/(b*Zh_Kdp_lut*Zh_Kdp_lut)));
    }
    dAh_dlna[i][i] = 0.5 * ((dAh_dlna[i+1][i]
     = -2.0*drange*dZh_Ah_dln_Zh_R*Zh_unat/(b*Zh_Ah_lut*Zh_Ah_lut)));
    dAv_dlna[i][i] = 0.5 * ((dAv_dlna[i+1][i]
     = -2.0*drange*dZh_Av_dln_Zh_R*Zh_unat/(b*Zh_Av_lut*Zh_Av_lut)));
    if (dZdr_dlna) {
      dZdr_dlna[i][i] = dZdr_dln_Zh_R/b;
    }
    /* Upper triangular elements */
    if (dphi_dp_dlna) {
      for (j = 0; j < i; j++) {
	dphi_dp_dlna[i+1][j] = dphi_dp_dlna[i][j]
	  + dAh_dlna[i][j] * dKdp_dZh * factor;
      }
    }
    if (dZdr_dlna) {
      for (j = 0; j < i; j++) {      
	dAh_dlna[i+1][j] = dAh_dlna[i][j] * (1.0 + factor*dAh_dZh);
	dAv_dlna[i+1][j] = dAv_dlna[i][j] * (1.0 + factor*dAv_dZh);
	dZdr_dlna[i+1][j] = - dAh_dlna[i+1][j] + dAv_dlna[i+1][j];
      } 
    }
  }

  return OE_SUCCESS;
}
  
  
