#include <mex.h>
#include "zdr_phidp.h"

int
get_string(const mxArray *m_array, char **c_string, char *name)
{
  if (mxIsChar(m_array) != 1) {
    mexPrintf("Reading %s\n", name);
    mexErrMsgTxt("Argument must be a string");
  }
  else {
    int buflen = mxGetM(m_array) * mxGetN(m_array) + 1;
    *c_string = (char *) mxCalloc(buflen, sizeof(char));
    mxGetString(m_array, *c_string, buflen);
    return buflen;
  }
}

int
get_double(const mxArray *m_array, OEreal *c_value, char *name)
{
  if (mxIsDouble(m_array) != 1) {
    mexPrintf("Reading %s\n", name);
    mexErrMsgTxt("Argument must be a double-precision value");
  }
  else {
    *c_value = *(mxGetPr(m_array));
    return mxGetM(m_array) * mxGetN(m_array);
  }
}

int
get_vector(const mxArray *m_array, OEreal **c_array, int n, char *name)
{
  if (mxIsDouble(m_array) != 1) {
    mexPrintf("Reading %s\n", name);
    mexErrMsgTxt("Argument must be a double-precision vector");
  }
  else {
    int the_size = mxGetM(m_array) * mxGetN(m_array);
    if (n > 0 && the_size != n) {
      mexPrintf("Argument %s has %d elements but should have %d\n",
		name, the_size, n);
      mexErrMsgTxt("Size mismatch");
    }
    *c_array = mxGetPr(m_array);
    return the_size;
  }
}

OEreal**
create_matrix_pointers(const mxArray *m_array)
{
  int m = mxGetM(m_array);
  int n = mxGetN(m_array);
  OEreal **ans = mxCalloc(m, sizeof(OEreal*));
  OEreal *ptr = mxGetPr(m_array);
  int i;
  for (i = 0; i < m; i++) {
    ans[i] = ptr + i*n;
  }
  return ans;
}


void
mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
  /* Forward model input */
  char *lut_filename;
  int n;
  OEreal drange;
  OEreal *range;
  OEreal *temperature;
  OEreal *Zh_lin;
  OEreal *a;
  OEreal b;
  /* Forward model output */
  OEreal *Zdr;
  OEreal *phi_dp;
  OEreal *Ah;
  OEreal *Av;
  OEreal **dZdr_dlna;
  OEreal **dphi_dp_dlna;
  OEreal **dAh_dlna;
  OEreal **dAv_dlna;

  if (nrhs != 7) {
    mexErrMsgTxt("Must have exactly 7 input arguments");
  }

  if (nlhs != 8) {
    mexErrMsgTxt("Must have exactly 8 output arguments");
  }

  get_string(prhs[0], &lut_filename, "scattering filename");
  get_double(prhs[1], &drange, "range-gate spacing");
  n = get_vector(prhs[2], &range, -1, "range");
  get_vector(prhs[3], &temperature, n, "temperature");
  get_vector(prhs[4], &Zh_lin, n, "linear reflectivity");
  get_vector(prhs[5], &a, n, "a");
  get_double(prhs[6], &b, "b");
  
  plhs[0] = mxCreateDoubleMatrix(n, 1, mxREAL);
  plhs[1] = mxCreateDoubleMatrix(n, 1, mxREAL);
  plhs[2] = mxCreateDoubleMatrix(n, 1, mxREAL);
  plhs[3] = mxCreateDoubleMatrix(n, 1, mxREAL);
  plhs[4] = mxCreateDoubleMatrix(n, n, mxREAL);
  plhs[5] = mxCreateDoubleMatrix(n, n, mxREAL);
  plhs[6] = mxCreateDoubleMatrix(n, n, mxREAL);
  plhs[7] = mxCreateDoubleMatrix(n, n, mxREAL);

  Zdr = mxGetPr(plhs[0]);
  phi_dp = mxGetPr(plhs[1]);
  Ah = mxGetPr(plhs[2]);
  Av = mxGetPr(plhs[3]);

  dZdr_dlna = create_matrix_pointers(plhs[4]);
  dphi_dp_dlna = create_matrix_pointers(plhs[5]);
  dAh_dlna = create_matrix_pointers(plhs[6]);
  dAv_dlna = create_matrix_pointers(plhs[7]);

  oe_set_memory_functions(mxMalloc, mxRealloc, mxFree);

  if (forward_model(lut_filename, n, drange, range,
		    temperature, Zh_lin, a, b,
		    /* Output */
		    Zdr, phi_dp, Ah, Av,
		    dZdr_dlna, dphi_dp_dlna, dAh_dlna, dAv_dlna)) {
    mexErrMsgTxt("An error occurred when calling the forward model function");
  }

  mxFree(dZdr_dlna);
  mxFree(dphi_dp_dlna);
  mxFree(dAh_dlna);
  mxFree(dAv_dlna);
}
